# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2015, Thibault Saunier <tsaunier@gnome.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import os
import time
from fractions import Fraction
from gettext import gettext as _

from gi.repository import GES
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstPbutils
from gi.repository import GstTranscoder

from pitivi.check import GstDependency
from pitivi.configure import get_gstpresets_dir
from pitivi.dialogs.prefs import PreferencesDialog
from pitivi.settings import GlobalSettings
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import ASSET_DURATION_META
from pitivi.utils.misc import asset_get_duration
# Remove check when we depend on Gst >= 1.20
HAS_GST_1_19 = GstDependency("Gst", apiversion="1.0", version_required="1.19").check()

# Make sure gst knowns about our own GstPresets
Gst.preset_set_app_dir(get_gstpresets_dir())


class ProxyingStrategy:
    AUTOMATIC = "automatic"
    ALL = "all"
    NOTHING = "nothing"


GlobalSettings.add_config_section("proxy")
GlobalSettings.add_config_option('proxying_strategy',
                                 section='proxy',
                                 key='proxying-strategy',
                                 default=ProxyingStrategy.AUTOMATIC)

GlobalSettings.add_config_option('num_transcoding_jobs',
                                 section='proxy',
                                 key='num-proxying-jobs',
                                 default=4,
                                 notify=True)
PreferencesDialog.add_numeric_preference('num_transcoding_jobs',
                                         description="",
                                         section="_proxies",
                                         label=_("Max number of parallel transcoding jobs"),
                                         lower=1)

GlobalSettings.add_config_option("max_cpu_usage",
                                 section="proxy",
                                 key="max-cpu-usage",
                                 default=10,
                                 notify=True)
PreferencesDialog.add_numeric_preference('max_cpu_usage',
                                         description="",
                                         section="_proxies",
                                         label=_("Max CPU usage dedicated to transcoding"),
                                         lower=1,
                                         upper=100)


GlobalSettings.add_config_option("auto_scaling_enabled",
                                 section="proxy",
                                 key="s-proxy-enabled",
                                 default=False,
                                 notify=True)
GlobalSettings.add_config_option("default_scaled_proxy_width",
                                 section="proxy",
                                 key="s-proxy-width",
                                 default=1920,
                                 notify=True)
GlobalSettings.add_config_option("default_scaled_proxy_height",
                                 section="proxy",
                                 key="s-proxy-height",
                                 default=1080,
                                 notify=True)

ENCODING_FORMAT_PRORES = "prores-raw-in-qt.gep"
ENCODING_FORMAT_JPEG = "jpeg-raw-in-qt.gep"


def create_encoding_profile_simple(container_caps, audio_caps, video_caps):
    container_profile = GstPbutils.EncodingContainerProfile.new(None, None,
                                                                Gst.Caps(container_caps),
                                                                None)
    audio_profile = GstPbutils.EncodingAudioProfile.new(Gst.Caps(audio_caps),
                                                        None, None, 0)
    video_profile = GstPbutils.EncodingVideoProfile.new(Gst.Caps(video_caps),
                                                        None, None, 0)

    container_profile.add_profile(audio_profile)
    container_profile.add_profile(video_profile)

    return container_profile


class ProxyManager(GObject.Object, Loggable):
    """Transcodes assets and manages proxies."""

    __gsignals__ = {
        "progress": (GObject.SignalFlags.RUN_LAST, None, (object, int, int)),
        "proxy-ready": (GObject.SignalFlags.RUN_LAST, None, (object, object)),
        "asset-preparing-cancelled": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "error-preparing-asset": (GObject.SignalFlags.RUN_LAST, None, (object, object, object)),
    }

    WHITELIST_CONTAINER_CAPS = ["video/quicktime", "application/ogg", "application/xges",
                                "video/x-matroska", "video/webm", "image/jpeg"]
    WHITELIST_AUDIO_CAPS = ["audio/mpeg", "audio/x-vorbis",
                            "audio/x-raw", "audio/x-flac",
                            "audio/x-wav"]
    WHITELIST_VIDEO_CAPS = ["video/x-h264", "image/jpeg",
                            "video/x-raw", "video/x-vp8",
                            "video/x-theora"]

    WHITELIST_FORMATS = []
    for container in WHITELIST_CONTAINER_CAPS:
        for audio in WHITELIST_AUDIO_CAPS:
            for video in WHITELIST_VIDEO_CAPS:
                WHITELIST_FORMATS.append(create_encoding_profile_simple(
                    container, audio, video))

    for audio in WHITELIST_AUDIO_CAPS:
        a = GstPbutils.EncodingAudioProfile.new(Gst.Caps(audio), None, None, 0)
        WHITELIST_FORMATS.append(a)

    hq_proxy_extension = "proxy.mov"
    scaled_proxy_extension = "scaledproxy.mov"
    # Suffix for filenames of proxies being created.
    part_suffix = ".part"

    def __init__(self, app):
        GObject.Object.__init__(self)
        Loggable.__init__(self)

        self.app = app
        # Total time to transcode in seconds.
        self._total_time_to_transcode = 0
        # Transcoded time per asset in seconds.
        self._transcoded_durations = {}
        self._start_proxying_time = 0
        self.__running_transcoders = []
        self.__pending_transcoders = []
        # The scaled proxy transcoders waiting for their corresponding shadow
        # HQ proxy transcoder to finish.
        self.__waiting_transcoders = []

        self.__encoding_target_file = None
        self.proxying_unsupported = False
        for encoding_format in [ENCODING_FORMAT_JPEG, ENCODING_FORMAT_PRORES]:
            self.__encoding_profile = self.__get_encoding_profile(encoding_format)
            if self.__encoding_profile:
                self.__encoding_target_file = encoding_format
                self.info("Using %s as proxying format", encoding_format)
                break

        if not self.__encoding_profile:
            self.proxying_unsupported = True

            self.error("Not supporting any proxy formats!")
            return

    def _scale_asset_resolution(self, asset, max_width, max_height):
        stream = asset.get_info().get_video_streams()[0]
        width = stream.get_width()
        height = stream.get_height()
        aspect_ratio = Fraction(width, height)

        if aspect_ratio.numerator >= width or aspect_ratio.denominator >= height:
            self.log("Unscalable aspect ratio.")
            return width, height
        if aspect_ratio.numerator >= max_width or aspect_ratio.denominator >= max_height:
            self.log("Cannot scale to target resolution.")
            return width, height

        if width > max_width or height > max_height:
            width_factor = max_width // aspect_ratio.numerator
            height_factor = max_height // aspect_ratio.denominator
            scaling_factor = min(height_factor, width_factor)

            width = aspect_ratio.numerator * scaling_factor
            height = aspect_ratio.denominator * scaling_factor

        return width, height

    def _asset_matches_encoding_format(self, asset, encoding_profile):
        def caps_match(info, profile):
            return not info.get_caps().intersect(profile.get_format()).is_empty()

        info = asset.get_info()
        if isinstance(encoding_profile, GstPbutils.EncodingAudioProfile):
            if isinstance(info.get_stream_info(), GstPbutils.DiscovererContainerInfo):
                return False
            audios = info.get_audio_streams()
            if len(audios) != 1 or not caps_match(audios[0], encoding_profile):
                return False
            if info.get_video_streams():
                return False
            return True

        container = info.get_stream_info()
        if container:
            if not caps_match(container, encoding_profile):
                return False

        for profile in encoding_profile.get_profiles():
            if isinstance(profile, GstPbutils.EncodingAudioProfile):
                audios = info.get_audio_streams()
                for audio_stream in audios:
                    if not caps_match(audio_stream, profile):
                        return False
            elif isinstance(profile, GstPbutils.EncodingVideoProfile):
                videos = info.get_video_streams()
                for video_stream in videos:
                    if not caps_match(video_stream, profile):
                        return False
        return True

    def __get_encoding_profile(self, encoding_target_file, asset=None, width=None,
                               height=None):
        encoding_target = GstPbutils.EncodingTarget.load_from_file(
            os.path.join(get_gstpresets_dir(), encoding_target_file))
        encoding_profile = encoding_target.get_profile("default")

        if not encoding_profile:
            return None

        for profile in encoding_profile.get_profiles():
            profile_format = profile.get_format()
            # Do not verify we have an encoder/decoder for raw audio/video,
            # as they are not required.
            if profile_format.intersect(Gst.Caps("audio/x-raw(ANY)")) or \
                    profile_format.intersect(Gst.Caps("audio/x-video(ANY)")):
                continue
            if not Gst.ElementFactory.list_filter(
                    Gst.ElementFactory.list_get_elements(
                        Gst.ELEMENT_FACTORY_TYPE_ENCODER, Gst.Rank.MARGINAL),
                    profile_format, Gst.PadDirection.SRC, False):
                return None
            if height and width and profile.get_type_nick() == "video":
                profile.set_restriction(Gst.Caps.from_string(
                    "video/x-raw, width=%d, height=%d" % (width, height)))

            if not Gst.ElementFactory.list_filter(
                    Gst.ElementFactory.list_get_elements(
                        Gst.ELEMENT_FACTORY_TYPE_DECODER, Gst.Rank.MARGINAL),
                    profile_format, Gst.PadDirection.SINK, False):
                return None

        if asset:
            # If we have an asset, we force audioconvert to keep
            # the number of channels
            # TODO: remove once https://bugzilla.gnome.org/show_bug.cgi?id=767226
            # is fixed
            info = asset.get_info()
            try:
                # TODO Be smarter about multiple streams
                audio_stream = info.get_audio_streams()[0]
                channels = audio_stream.get_channels()
                audio_profile = [
                    profile for profile in encoding_profile.get_profiles()
                    if isinstance(profile, GstPbutils.EncodingAudioProfile)][0]
                audio_profile.set_restriction(Gst.Caps.from_string(
                    "audio/x-raw,channels=%d" % channels))
            except IndexError:
                pass

        return encoding_profile

    @classmethod
    def is_proxy_asset(cls, obj):
        return cls.is_scaled_proxy(obj) or cls.is_hq_proxy(obj)

    @classmethod
    def is_scaled_proxy(cls, obj):
        if isinstance(obj, GES.Asset):
            uri = obj.props.id
        else:
            uri = obj

        return uri.endswith("." + cls.scaled_proxy_extension)

    @classmethod
    def is_hq_proxy(cls, obj):
        if isinstance(obj, GES.Asset):
            uri = obj.props.id
        else:
            uri = obj

        return uri.endswith("." + cls.hq_proxy_extension)

    def check_proxy_loading_succeeded(self, proxy):
        if self.is_proxy_asset(proxy):
            return True

        self.emit("error-preparing-asset", None, proxy, proxy.get_error())
        return False

    @classmethod
    def get_target_uri(cls, obj):
        if isinstance(obj, GES.Asset):
            uri = obj.props.id
        else:
            uri = obj

        if cls.is_scaled_proxy(uri):
            return ".".join(uri.split(".")[:-4])

        if cls.is_proxy_asset(uri):
            return ".".join(uri.split(".")[:-3])

        return uri

    def get_proxy_uri(self, asset, scaled=False):
        """Gets the URI of the corresponding proxy file for the specified asset.

        The name looks like:
            <filename>.<file_size>[.<proxy_resolution>].<proxy_extension>

        Returns:
            str: The URI or None if it can't be computed for any reason.
        """
        asset_file = Gio.File.new_for_uri(asset.get_id())
        try:
            file_size = asset_file.query_info(Gio.FILE_ATTRIBUTE_STANDARD_SIZE,
                                              Gio.FileQueryInfoFlags.NONE,
                                              None).get_size()
        except GLib.Error as err:
            if err.matches(Gio.io_error_quark(), Gio.IOErrorEnum.NOT_FOUND):
                return None
            else:
                raise

        if scaled:
            if not asset.get_info().get_video_streams():
                return None

            max_w = self.app.project_manager.current_project.scaled_proxy_width
            max_h = self.app.project_manager.current_project.scaled_proxy_height
            t_width, t_height = self._scale_asset_resolution(asset, max_w, max_h)
            proxy_res = "%sx%s" % (t_width, t_height)
            return "%s.%s.%s.%s" % (asset.get_id(), file_size, proxy_res,
                                    self.scaled_proxy_extension)
        else:
            return "%s.%s.%s" % (asset.get_id(), file_size,
                                 self.hq_proxy_extension)

    def is_asset_format_well_supported(self, asset):
        for encoding_format in self.WHITELIST_FORMATS:
            if self._asset_matches_encoding_format(asset, encoding_format):
                self.info("Automatically not proxying")
                return True

        return False

    def asset_matches_target_res(self, asset):
        """Returns whether the asset's size <= the scaled proxy size."""
        stream = asset.get_info().get_video_streams()[0]

        asset_res = (stream.get_width(), stream.get_height())
        target_res = self._scale_asset_resolution(asset,
                                                  self.app.project_manager.current_project.scaled_proxy_width,
                                                  self.app.project_manager.current_project.scaled_proxy_height)

        return asset_res == target_res

    def __asset_needs_transcoding(self, asset, scaled=False):
        if self.proxying_unsupported:
            self.info("No proxying supported")
            return False

        if asset.is_image():
            return False

        if self.app.settings.proxying_strategy == ProxyingStrategy.NOTHING:
            self.debug("Not proxying anything. %s",
                       self.app.settings.proxying_strategy)
            return False

        if self.app.settings.proxying_strategy == ProxyingStrategy.AUTOMATIC \
                and scaled and not self.asset_matches_target_res(asset):
            return True

        if self.app.settings.proxying_strategy == ProxyingStrategy.AUTOMATIC \
                and not scaled and not self.is_hq_proxy(asset) and \
                self.is_asset_format_well_supported(asset):
            return False

        if not self._asset_matches_encoding_format(asset, self.__encoding_profile):
            return True

        self.info("%s does not need proxy", asset.get_id())
        return False

    def asset_can_be_proxied(self, asset, scaled=False):
        """Returns whether the asset is not a proxy nor a proper proxy."""
        if asset.is_image():
            return False

        if scaled:
            if not asset.get_info().get_video_streams():
                return False

            return not self.is_scaled_proxy(asset) or \
                self.asset_matches_target_res(asset)
        else:
            return not self.is_hq_proxy(asset)

    def __start_transcoder(self, transcoder):
        self.debug("Starting %s", transcoder.props.src_uri)
        if self._start_proxying_time == 0:
            self._start_proxying_time = time.time()
        transcoder.run_async()
        self.__running_transcoders.append(transcoder)

    def __assets_match(self, asset, proxy):
        if self.__asset_needs_transcoding(proxy):
            return False

        info = asset.get_info()
        if info.get_duration() != asset.get_duration():
            return False

        return True

    def __asset_loaded_cb(self, proxy, res, asset, transcoder):
        try:
            GES.Asset.request_finish(res)
        except GLib.Error as e:
            if transcoder:
                self.emit("error-preparing-asset", asset, proxy, e)
                del transcoder
            else:
                self.__create_transcoder(asset)

            return

        shadow = transcoder and self._is_shadow_transcoder(transcoder)

        if not transcoder:
            if not self.__assets_match(asset, proxy):
                self.__create_transcoder(asset)
                return
        else:
            if transcoder.props.pipeline.props.video_filter:
                transcoder.props.pipeline.props.video_filter.finalize()

            if transcoder.props.pipeline.props.audio_filter:
                transcoder.props.pipeline.props.audio_filter.finalize()

            del transcoder

        asset_duration = asset_get_duration(asset)
        proxy_duration = asset_get_duration(proxy)
        if asset_duration != proxy_duration:
            duration = min(asset_duration, proxy_duration)
            self.info("Resetting %s duration from %s to %s as"
                      " new proxy has a different duration",
                      asset.props.id, Gst.TIME_ARGS(asset_duration), Gst.TIME_ARGS(duration))
            asset.set_uint64(ASSET_DURATION_META, duration)
            proxy.set_uint64(ASSET_DURATION_META, duration)
            target_uri = self.get_target_uri(asset)

            for clip in self.app.project_manager.current_project.ges_timeline.iter_clips():
                if not isinstance(clip, GES.UriClip):
                    continue
                if self.get_target_uri(clip.props.uri) == target_uri:
                    if clip.props.in_point + clip.props.duration > duration:
                        new_duration = duration - clip.props.in_point
                        if new_duration > 0:
                            self.warning("%s resetting duration to %s as"
                                         " new proxy has a shorter duration",
                                         clip, Gst.TIME_ARGS(new_duration))
                            clip.set_duration(new_duration)
                        else:
                            new_inpoint = new_duration - clip.props.in_point
                            self.error("%s resetting duration to %s"
                                       " and inpoint to %s as the proxy"
                                       " is shorter",
                                       clip, Gst.TIME_ARGS(new_duration), Gst.TIME_ARGS(new_inpoint))
                            clip.set_inpoint(new_inpoint)
                            clip.set_duration(duration - new_inpoint)
                        clip.set_max_duration(duration)

        if shadow:
            self.app.project_manager.current_project.finalize_proxy(proxy)
        else:
            self.emit("proxy-ready", asset, proxy)
            self.__emit_progress(proxy, 100)

    def __transcoder_error_cb(self, _, error, unused_details, asset, transcoder):
        self.emit("error-preparing-asset", asset, None, error)

    def __transcoder_done_cb(self, emitter, asset, transcoder):
        emitter.disconnect_by_func(self.__proxying_position_changed_cb)
        emitter.disconnect_by_func(self.__transcoder_done_cb)
        emitter.disconnect_by_func(self.__transcoder_error_cb)

        self.debug("Transcoder done with %s", asset.get_id())

        self.__running_transcoders.remove(transcoder)

        proxy_uri = transcoder.props.dest_uri.rstrip(ProxyManager.part_suffix)
        os.rename(Gst.uri_get_location(transcoder.props.dest_uri),
                  Gst.uri_get_location(proxy_uri))

        shadow = self._is_shadow_transcoder(transcoder)
        second_transcoder = self._get_second_transcoder(transcoder)
        if second_transcoder and not shadow:
            # second_transcoder is the shadow for transcoder.
            # Defer loading until the shadow transcoder finishes.
            self.__waiting_transcoders.append([transcoder, asset])
        else:
            # Make sure that if it first failed loading, the proxy is forced to
            # be reloaded in the GES cache.
            GES.Asset.needs_reload(GES.UriClip, proxy_uri)
            GES.Asset.request_async(GES.UriClip, proxy_uri, None,
                                    self.__asset_loaded_cb, asset, transcoder)

        if shadow:
            # Finish deferred loading for waiting scaled proxy transcoder.
            for pair in self.__waiting_transcoders:
                waiting_transcoder, waiting_asset = pair
                if waiting_transcoder.props.src_uri == transcoder.props.src_uri:
                    proxy_uri = waiting_transcoder.props.dest_uri.rstrip(ProxyManager.part_suffix)
                    GES.Asset.needs_reload(GES.UriClip, proxy_uri)
                    GES.Asset.request_async(GES.UriClip, proxy_uri, None,
                                            self.__asset_loaded_cb, waiting_asset, waiting_transcoder)

                    self.__waiting_transcoders.remove(pair)
                    break

        try:
            self.__start_transcoder(self.__pending_transcoders.pop())
        except IndexError:
            if not self.__running_transcoders:
                self._transcoded_durations = {}
                self._total_time_to_transcode = 0
                self._start_proxying_time = 0

    def __emit_progress(self, asset, creation_progress):
        """Handles the transcoding progress of the specified asset."""
        if self._transcoded_durations:
            time_spent = time.time() - self._start_proxying_time
            transcoded_seconds = sum(self._transcoded_durations.values())
            remaining_seconds = max(0, self._total_time_to_transcode - transcoded_seconds)
            estimated_time = remaining_seconds * time_spent / transcoded_seconds
        else:
            estimated_time = 0

        asset.creation_progress = creation_progress
        self.emit("progress", asset, asset.creation_progress, estimated_time)

    def __proxying_position_changed_cb(self, _, position, asset, transcoder):
        if transcoder not in self.__running_transcoders:
            self.info("Position changed after job cancelled!")
            return

        second_transcoder = self._get_second_transcoder(transcoder)
        if second_transcoder is not None:
            position = (position + second_transcoder.props.position) // 2

        self._transcoded_durations[asset] = position / Gst.SECOND

        duration = transcoder.props.duration
        if duration <= 0 or duration == Gst.CLOCK_TIME_NONE:
            duration = asset.props.duration
        if duration > 0 and duration != Gst.CLOCK_TIME_NONE:
            creation_progress = 100 * position / duration
            # Do not set to >= 100 as we need to notify about the proxy first.

            asset.creation_progress = max(0, min(creation_progress, 99))

        self.__emit_progress(asset, asset.creation_progress)

    def _get_second_transcoder(self, transcoder):
        """Gets the shadow of a scaled proxy or the other way around."""
        all_transcoders = self.__running_transcoders + self.__pending_transcoders
        for transcoder2 in all_transcoders:
            if transcoder2.props.position_update_interval == transcoder.props.position_update_interval:
                # Both transcoders are of the same type.
                continue
            if transcoder2.props.src_uri == transcoder.props.src_uri:
                return transcoder2
        return None

    def _is_shadow_transcoder(self, transcoder):
        if transcoder.props.position_update_interval == 1001:
            return True
        return False

    def is_asset_queued(self, asset, optimisation=True, scaling=True):
        """Returns whether the specified asset is queued for transcoding.

        Args:
            asset (GES.Asset): The asset to check.
            optimisation(bool): Whether to check optimisation queue
            scaling(bool): Whether to check scaling queue

        Returns:
            bool: True if the asset is being transcoded or pending.
        """
        all_transcoders = self.__running_transcoders + self.__pending_transcoders
        is_queued = False
        for transcoder in all_transcoders:
            transcoder_uri = transcoder.props.dest_uri
            scaling_ext = "." + self.scaled_proxy_extension + ProxyManager.part_suffix
            optimisation_ext = "." + self.hq_proxy_extension + ProxyManager.part_suffix

            scaling_transcoder = transcoder_uri.endswith(scaling_ext)
            optimisation_transcoder = transcoder_uri.endswith(optimisation_ext)

            if transcoder.props.src_uri == asset.props.id:
                if optimisation and optimisation_transcoder:
                    is_queued = True
                    break

                if scaling and scaling_transcoder:
                    is_queued = True
                    break

        return is_queued

    def __create_transcoder(self, asset, scaled=False, shadow=False):
        self._total_time_to_transcode += asset.get_duration() / Gst.SECOND
        asset_uri = asset.get_id()
        proxy_uri = self.get_proxy_uri(asset, scaled=scaled)

        if Gio.File.new_for_uri(proxy_uri).query_exists(None):
            self.debug("Using proxy already generated: %s", proxy_uri)
            GES.Asset.request_async(GES.UriClip,
                                    proxy_uri, None,
                                    self.__asset_loaded_cb, asset,
                                    None)
            return

        self.debug("Creating a proxy for %s (strategy: %s, force: %s, scaled: %s)",
                   asset.get_id(), self.app.settings.proxying_strategy,
                   asset.force_proxying, scaled)

        width = None
        height = None
        if scaled:
            project = self.app.project_manager.current_project
            w = project.scaled_proxy_width
            h = project.scaled_proxy_height
            if not project.has_scaled_proxy_size():
                project.scaled_proxy_width = w
                project.scaled_proxy_height = h
            width, height = self._scale_asset_resolution(asset, w, h)
        enc_profile = self.__get_encoding_profile(self.__encoding_target_file,
                                                  asset, width, height)

        if HAS_GST_1_19:
            transcoder = GstTranscoder.Transcoder.new_full(
                asset_uri, proxy_uri + ProxyManager.part_suffix, enc_profile)
            signals_emitter = transcoder.get_signal_adapter(None)
        else:
            dispatcher = GstTranscoder.TranscoderGMainContextSignalDispatcher.new()
            signals_emitter = transcoder = GstTranscoder.Transcoder.new_full(
                asset_uri, proxy_uri + ProxyManager.part_suffix, enc_profile,
                dispatcher)

        if shadow:
            # Used to identify shadow transcoder
            transcoder.props.position_update_interval = 1001
        else:
            transcoder.props.position_update_interval = 1000

        info = asset.get_info()
        if info.get_video_streams():
            thumbnailbin = Gst.ElementFactory.make("teedthumbnailbin")
            thumbnailbin.props.uri = asset.get_id()
            transcoder.props.pipeline.props.video_filter = thumbnailbin

        if info.get_audio_streams():
            waveformbin = Gst.ElementFactory.make("waveformbin")
            waveformbin.props.uri = asset.get_id()
            waveformbin.props.duration = asset.get_duration()
            transcoder.props.pipeline.props.audio_filter = waveformbin

        transcoder.set_cpu_usage(self.app.settings.max_cpu_usage)
        signals_emitter.connect("position-updated",
                                self.__proxying_position_changed_cb,
                                asset, transcoder)

        signals_emitter.connect("done", self.__transcoder_done_cb, asset, transcoder)
        signals_emitter.connect("error", self.__transcoder_error_cb, asset, transcoder)

        if len(self.__running_transcoders) < self.app.settings.num_transcoding_jobs:
            self.__start_transcoder(transcoder)
        else:
            self.__pending_transcoders.append(transcoder)

    def cancel_job(self, asset):
        """Cancels the transcoding job for the specified asset, if any.

        Args:
            asset (GES.Asset): The original asset.
        """
        if not self.is_asset_queued(asset):
            return

        for transcoder in self.__running_transcoders:
            if asset.props.id == transcoder.props.src_uri:
                self.info("Cancelling running transcoder %s %s",
                          transcoder.props.src_uri,
                          transcoder.__grefcount__)
                self.__running_transcoders.remove(transcoder)
                self.emit("asset-preparing-cancelled", asset)

        for transcoder in self.__pending_transcoders:
            if asset.props.id == transcoder.props.src_uri:
                self.info("Cancelling pending transcoder %s",
                          transcoder.props.src_uri)
                # Removing the transcoder from the list
                # will lead to its destruction (only reference)
                # here, which means it will be stopped.
                self.__pending_transcoders.remove(transcoder)
                self.emit("asset-preparing-cancelled", asset)

    def add_job(self, asset, scaled=False, shadow=False):
        """Adds a transcoding job for the specified asset if needed.

        Args:
            asset (GES.Asset): The asset to be transcoded.
            scaled (Optional[bool]): Whether to create a scaled proxy instead
                of a high-quality proxy.
            shadow (Optional[bool]): Whether to create a high-quality proxy
                to shadow a scaled proxy.
        """
        force_proxying = asset.force_proxying
        video_streams = asset.get_info().get_video_streams()
        if video_streams:
            # Handle Automatic scaling
            if self.app.settings.auto_scaling_enabled and not force_proxying \
                    and not shadow and not self.asset_matches_target_res(asset):
                scaled = True

            # Create shadow proxies for unsupported assets
            if not self.is_asset_format_well_supported(asset) and not \
                    self.app.settings.proxying_strategy == ProxyingStrategy.NOTHING \
                    and not shadow and scaled:
                hq_uri = self.app.proxy_manager.get_proxy_uri(asset)
                if not Gio.File.new_for_uri(hq_uri).query_exists(None):
                    self.add_job(asset, shadow=True)
        else:
            # Scaled proxy is not for audio assets
            scaled = False

        if self.is_asset_queued(asset, scaling=scaled, optimisation=not scaled):
            self.log("Asset %s already queued for %s", asset, "scaling" if scaled else "optimization")
            return

        if not force_proxying:
            if not self.__asset_needs_transcoding(asset, scaled):
                self.debug("Not proxying asset (proxying disabled: %s)",
                           self.proxying_unsupported)
                # Make sure to notify we do not need a proxy for that asset.
                self.emit("proxy-ready", asset, None)
                return

        self.__create_transcoder(asset, scaled=scaled, shadow=shadow)


def get_proxy_target(obj):
    if isinstance(obj, GES.UriClip):
        asset = obj.get_asset()
    elif isinstance(obj, GES.TrackElement):
        asset = obj.get_parent().get_asset()
    else:
        asset = obj

    if ProxyManager.is_proxy_asset(asset):
        target = asset.get_proxy_target()
        if target and target.get_error() is None:
            asset = target

    return asset



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (C) 2020 Andrew Hazel, Thomas Braccia, Troy Ogden, Robert Kirkpatrick
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Widgets to control clips properties."""
import os
from gettext import gettext as _

from gi.repository import GES
from gi.repository import Gtk

from pitivi.configure import get_ui_dir
from pitivi.settings import GlobalSettings
from pitivi.undo.timeline import CommitTimelineFinalizingAction
from pitivi.utils.loggable import Loggable
from pitivi.utils.ui import argb_to_gdk_rgba
from pitivi.utils.ui import gdk_rgba_to_argb
from pitivi.utils.widgets import ColorPickerButton


GlobalSettings.add_config_section("user-interface")

GlobalSettings.add_config_option("ColorClipLength",
                                 section="user-interface",
                                 key="color-clip-length",
                                 default=5000,
                                 notify=True)


class ColorProperties(Gtk.Expander, Loggable):
    """Widget for configuring the properties of a color clip."""

    def __init__(self, app):
        Gtk.Expander.__init__(self)
        Loggable.__init__(self)

        self.app = app
        self.source = None
        self._children_props_handler = None

        self.set_label(_("Color"))
        self.set_expanded(True)

        self._create_ui()

    def _create_ui(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(os.path.join(get_ui_dir(), "clipcolor.ui"))
        self.builder.connect_signals(self)

        box = self.builder.get_object("color_box")
        self.add(box)

        self.color_button = self.builder.get_object("color_button")

        self.color_picker_button = ColorPickerButton()
        box.add(self.color_picker_button)
        self.color_picker_button.connect(
            "value-changed", self._color_picker_value_changed_cb)

        self.show_all()

    def _set_child_property(self, name, value):
        with self.app.action_log.started("Color change property",
                                         finalizing_action=CommitTimelineFinalizingAction(self.app.project_manager.current_project.pipeline),
                                         toplevel=True):
            res = self.source.set_child_property(name, value)
            assert res

    def _color_picker_value_changed_cb(self, widget):
        argb = widget.calculate_argb()
        self._set_child_property("foreground-color", argb)

    def _color_button_cb(self, widget):
        argb = gdk_rgba_to_argb(widget.get_rgba())
        self._set_child_property("foreground-color", argb)

    def set_source(self, source):
        """Sets the clip source to be edited with this editor.

        Args:
            source (GES.VideoTestSource): The source of the clip.
        """
        self.debug("Source set to %s", source)
        if self._children_props_handler is not None:
            self.source.disconnect(self._children_props_handler)
            self._children_props_handler = None
        self.source = None
        if source:
            assert isinstance(source, GES.VideoTestSource)
            self.source = source
            self._update_color_button()
            self._children_props_handler = self.source.connect("deep-notify",
                                                               self._source_deep_notify_cb)
        self.set_visible(bool(self.source))

    def _source_deep_notify_cb(self, source, unused_gstelement, pspec):
        """Handles updates in the VideoTestSource backing the current TestClip."""
        if pspec.name == "foreground-color":
            self._update_color_button()

    def _update_color_button(self):
        res, argb = self.source.get_child_property("foreground-color")
        assert res
        color = argb_to_gdk_rgba(argb)
        self.color_button.set_rgba(color)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2006, Edward Hervey <bilboed@bilboed.com>
# Copyright (c) 2014, Alex Băluț <alexandru.balut@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from gettext import gettext as _

import cairo
from gi.repository import Gdk
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.utils.loggable import Loggable
from pitivi.utils.timeline import Zoomable
from pitivi.utils.ui import beautify_length
from pitivi.utils.ui import FONT_SCALING_FACTOR
from pitivi.utils.ui import gtk_style_context_get_color
from pitivi.utils.ui import NORMAL_FONT
from pitivi.utils.ui import PLAYHEAD_COLOR
from pitivi.utils.ui import PLAYHEAD_WIDTH
from pitivi.utils.ui import set_cairo_color
from pitivi.utils.ui import time_to_string


# Tuples of:
# - an interval duration in seconds for which a timestamp will be displayed
# - how the ticks should be displayed for this interval:
#   (tick interval duration in seconds, height ratio) tuples.
SCALES = (
    (0.1, ((0.1, 1.0), (0.05, .5), (0.01, .25))),
    (0.2, ((0.2, 1.0), (0.1, .5), (0.05, .25))),
    (0.5, ((0.5, 1.0), (0.1, .25))),

    # 1 second.
    (1, ((1, 1.0), (0.5, .5), (0.1, .25))),
    (2, ((2, 1.0), (1, .5), (0.5, .25))),
    (5, ((5, 1.0), (1, .25))),
    (10, ((10, 1.0), (5, .5), (1, .25))),
    (20, ((20, 1.0), (10, .5), (1, .25))),
    (30, ((30, 1.0), (10, .5), (1, .25))),

    # 1 minute.
    (60, ((60, 1.0), (30, .5), (15, .25))),
    # 2 minutes.
    (120, ((120, 1.0), (60, .5), (30, .25))),
    # 5 minutes.
    (300, ((300, 1.0), (60, .25))),
    # 10 minutes.
    (600, ((600, 1.0), (300, .5), (60, .25))),
    # 30 minutes.
    (1800, ((1800, 1.0), (900, .5), (450, .25))),

    # 1 hour.
    (3600, ((3600, 1.0), (1800, .75), (900, .5))),
)

# The minimum distance between adjacent ticks.
MIN_TICK_SPACING_PIXELS = 6

# For displaying the times a bit to the right.
TIMES_LEFT_MARGIN_PIXELS = 3

# The minimum width for a frame to be displayed.
FRAME_MIN_WIDTH_PIXELS = 5
# How short it should be.
FRAME_HEIGHT_PIXELS = 5

NORMAL_FONT_SIZE = FONT_SCALING_FACTOR * 13
SMALL_FONT_SIZE = FONT_SCALING_FACTOR * 11


class ScaleRuler(Gtk.DrawingArea, Loggable):
    """Widget for displaying the ruler.

    Displays a series of consecutive intervals. For each interval its beginning
    time is shown. If zoomed in enough, shows the frames in alternate colors.

    Attributes:
        zoom (pitivi.utils.timeline.Zoomable): Zoom controller.
        settings (pitivi.settings.GlobalSettings): The settings of the app.
        style_context (Gtk.StyleContext): The style context for drawing.
        _pipeline (pitivi.utils.pipeline.Pipeline): The pipeline of the project.
    """

    def __init__(self, zoom, settings, style_context):
        Gtk.DrawingArea.__init__(self)
        Loggable.__init__(self)
        self.log("Creating new ScaleRuler")

        self.zoom = zoom
        self.settings = settings
        self.style_context = style_context
        self._pipeline = None
        self.ges_timeline = None

        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK |
                        Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK |
                        Gdk.EventMask.SCROLL_MASK)

        self.pixbuf = None

        # all values are in pixels
        self.pixbuf_offset = 0
        self.pixbuf_offset_painted = 0

        self.position = 0  # In nanoseconds

        self._scales = [(interval, list(reversed(ticks))) for interval, ticks in SCALES]

        # Update colors when theme or color preferences change.
        Gtk.Settings.get_default().connect("notify::gtk-theme-name", self._update_colors_cb)
        Gtk.Settings.get_default().connect("notify::gtk-application-prefer-dark-theme", self._update_colors_cb)

    def set_pipeline(self, pipeline):
        self._pipeline = pipeline
        self.ges_timeline = pipeline.props.timeline
        self._pipeline.connect("position", self._pipeline_position_cb)

    def _pipeline_position_cb(self, unused_pipeline, position):
        self.position = position
        self.queue_draw()

    def _set_colors(self):
        # pylint: disable=attribute-defined-outside-init
        context = self.get_style_context()
        color_normal = gtk_style_context_get_color(context, Gtk.StateFlags.NORMAL)
        color_insensitive = gtk_style_context_get_color(context, Gtk.StateFlags.BACKDROP)
        self._color_normal = color_normal
        self._color_dimmed = Gdk.RGBA(
            *[(x * 3 + y * 2) / 5
              for x, y in ((color_normal.red, color_insensitive.red),
                           (color_normal.green, color_insensitive.green),
                           (color_normal.blue, color_insensitive.blue))])

        self._color_subtle = Gdk.RGBA(
            *[(x * 3 + y * 2) / 10
              for x, y in ((color_normal.red, color_insensitive.red),
                           (color_normal.green, color_insensitive.green),
                           (color_normal.blue, color_insensitive.blue))])

        # Two colors with high contrast.
        self._color_frame = gtk_style_context_get_color(context, Gtk.StateFlags.LINK)

    def _update_colors_cb(self, settings, gparams):
        self._set_colors()

# Gtk.Widget overrides

    def do_configure_event(self, unused_event):
        width = self.get_allocated_width()
        height = self.get_allocated_height()
        self.debug("Configuring, height %d, width %d", width, height)

        # Destroy previous buffer
        if self.pixbuf is not None:
            self.pixbuf.finish()
            self.pixbuf = None

        # Create a new buffer
        self.pixbuf = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)

        self._set_colors()
        return False

    def do_draw(self, context):
        if self.pixbuf is None:
            self.info("No buffer to paint")
            return False

        pixbuf = self.pixbuf

        # Draw on a temporary context and then copy everything.
        drawing_context = cairo.Context(pixbuf)
        self.draw_background(drawing_context)
        self.draw_ruler(drawing_context)
        self.draw_position(drawing_context)
        pixbuf.flush()

        context.set_source_surface(self.pixbuf, 0.0, 0.0)
        context.paint()

        return False

    def do_button_press_event(self, event):
        if not self._pipeline:
            return False

        button = event.button
        if button == 3 or (button == 1 and self.settings.leftClickAlsoSeeks):
            self.debug("button pressed at x:%d", event.x)
            position = self.pixel_to_ns(event.x + self.pixbuf_offset)
            self._pipeline.simple_seek(position)
            self.__set_tooltip_text(position, True)
        return False

    def do_button_release_event(self, event):
        button = event.button
        if button == 3 or (button == 1 and self.settings.leftClickAlsoSeeks):
            self.debug("button released at x:%d", event.x)
            position = self.pixel_to_ns(event.x + self.pixbuf_offset)
            self.__set_tooltip_text(position)
        return False

    def do_motion_notify_event(self, event):
        if not self._pipeline:
            return False

        position = self.pixel_to_ns(event.x + self.pixbuf_offset)

        seek_mask = Gdk.ModifierType.BUTTON3_MASK
        if self.settings.leftClickAlsoSeeks:
            seek_mask |= Gdk.ModifierType.BUTTON1_MASK

        seeking = event.state & seek_mask
        if seeking:
            self.debug("motion at event.x %d", event.x)
            self._pipeline.simple_seek(position)
        self.__set_tooltip_text(position, seeking)

        return False

    def __set_tooltip_text(self, position, seeking=False):
        """Updates the tooltip."""
        position = max(0, position)
        if seeking:
            timeline_duration = self.ges_timeline.props.duration
            position = min(position, timeline_duration)
        human_time = beautify_length(position)
        cur_frame = self.ges_timeline.get_frame_at(position) + 1
        self.set_tooltip_text(human_time + "\n" + _("Frame #%d") % cur_frame)

# Drawing methods

    def draw_background(self, context):
        width = context.get_target().get_width()
        height = context.get_target().get_height()
        Gtk.render_background(self.style_context, context, 0, 0, width, height)

    def draw_ruler(self, context):
        context.set_font_face(NORMAL_FONT)
        context.set_font_size(NORMAL_FONT_SIZE)

        spacing, interval_seconds, ticks = self._get_spacing(context)
        offset = self.pixbuf_offset % spacing
        self.draw_frame_boundaries(context)
        self.draw_ticks(context, offset, spacing, interval_seconds, ticks)
        self.draw_times(context, offset, spacing, interval_seconds)

    def _get_spacing(self, context):
        # The longest timestamp we display is 0:00:00 because
        # when we display millis, they are displayed by themselves.
        min_interval_width = context.text_extents("0:00:00")[2] * 1.3
        zoomratio = self.zoom.zoomratio
        for interval_seconds, ticks in self._scales:
            interval_width = interval_seconds * zoomratio
            if interval_width >= min_interval_width:
                return interval_width, interval_seconds, ticks
        raise Exception(
            "Failed to find an interval size for textwidth:%s, zoomratio:%s" %
            (min_interval_width, zoomratio))

    def draw_ticks(self, context, offset, spacing, interval_seconds, ticks):
        for tick_interval, height_ratio in ticks:
            count_per_interval = interval_seconds / tick_interval
            space = spacing / count_per_interval
            if space < MIN_TICK_SPACING_PIXELS:
                continue
            paintpos = 0.5 - offset

            color = (self._color_normal if height_ratio == 1
                     else self._color_subtle)

            set_cairo_color(context, color)
            while paintpos < context.get_target().get_width():
                self._draw_tick(context, paintpos, height_ratio)
                paintpos += space

    def _draw_tick(self, context, paintpos, height_ratio):
        # We need to use 0.5 pixel offsets to get a sharp 1 px line in cairo
        paintpos = int(paintpos - 0.5) + 0.5
        target_height = context.get_target().get_height()
        y = int(target_height * (1 - height_ratio))
        context.set_line_width(1)
        context.move_to(paintpos, y)
        context.line_to(paintpos, target_height)
        context.close_path()
        context.stroke()

    def draw_times(self, context, offset, spacing, interval_seconds):
        # figure out what the optimal offset is
        interval = int(Gst.SECOND * interval_seconds)
        current_time = self.zoom.pixel_to_ns(self.pixbuf_offset)
        paintpos = TIMES_LEFT_MARGIN_PIXELS
        if offset > 0:
            current_time = current_time - (current_time % interval) + interval
            paintpos += spacing - offset

        set_cairo_color(context, self._color_normal)
        y_bearing = context.text_extents("0")[1]

        def split(x):
            # Seven elements: h : mm : ss . mmm
            # Using negative indices because the first element (hour)
            # can have a variable length.
            return x[:-10], x[-10], x[-9:-7], x[-7], x[-6:-4], x[-4], x[-3:]

        previous = split(time_to_string(max(0, current_time - interval)))
        width = context.get_target().get_width()
        while paintpos < width:
            context.move_to(int(paintpos), 1 - y_bearing)
            current = split(time_to_string(int(current_time)))
            millis = current_time % Gst.SECOND > 0
            self._draw_time(context, current, previous, millis)
            previous = current
            paintpos += spacing
            current_time += interval

    def _draw_time(self, context, current, previous, millis):
        hour = int(current[0])
        for index, (element, previous_element) in enumerate(zip(current, previous)):
            if index <= 1 and not hour:
                # Don't draw hour if 0.
                continue
            if millis:
                # Draw only the millis.
                if index < 5:
                    continue
            else:
                # Don't draw the millis.
                if index == 5:
                    break
            if element == previous_element:
                color = self._color_dimmed
            else:
                color = self._color_normal
            set_cairo_color(context, color)
            # Display the millis with a smaller font.
            small = index >= 5
            if small:
                context.set_font_size(SMALL_FONT_SIZE)
            context.show_text(element)
            if small:
                context.set_font_size(NORMAL_FONT_SIZE)

    def draw_frame_boundaries(self, context):
        """Draws the alternating rectangles that represent the project frames.

        These are drawn only at high zoom levels.

        These are based on the project's framerate settings, not the actual
        frames on the assets.
        """
        if not self.ges_timeline:
            # Timeline not set yet
            return

        frame_width = self.zoom.ns_to_pixel(self.ges_timeline.get_frame_time(1))
        if frame_width < FRAME_MIN_WIDTH_PIXELS:
            return

        offset = self.pixbuf_offset % frame_width
        height = context.get_target().get_height()
        y = int(height - FRAME_HEIGHT_PIXELS)

        frame_num = self.ges_timeline.get_frame_at(self.zoom.pixel_to_ns(self.pixbuf_offset))
        paintpos = self.pixbuf_offset - offset
        max_pos = context.get_target().get_width() + self.pixbuf_offset
        while paintpos < max_pos:
            paintpos = self.zoom.ns_to_pixel(self.ges_timeline.get_frame_time(frame_num))
            if frame_num % 2:
                set_cairo_color(context, self._color_frame)
                context.rectangle(
                    0.5 + paintpos - self.pixbuf_offset, y, frame_width, height)
                context.fill()
            frame_num += 1

    def draw_position(self, context):
        """Draws the top part of the playhead.

        This should be in sync with the playhead drawn by the timeline.
        See Timeline.__draw_playhead().
        """
        height = self.pixbuf.get_height()

        semi_width = 4
        semi_height = int(semi_width * 1.61803)
        y = int(3 * height / 4)

        # Add 0.5 so that the line center is at the middle of the pixel,
        # without this the line appears blurry.
        xpos = self.zoom.ns_to_pixel(self.position) - self.pixbuf_offset + 0.5
        set_cairo_color(context, PLAYHEAD_COLOR)

        context.set_line_width(PLAYHEAD_WIDTH)
        context.move_to(xpos, y)
        context.line_to(xpos, height)
        context.stroke()

        context.set_line_width(PLAYHEAD_WIDTH * 2)
        context.move_to(xpos, y)
        context.line_to(xpos + semi_width, y - semi_height)
        context.line_to(xpos, y - semi_height * 2)
        context.line_to(xpos - semi_width, y - semi_height)
        context.close_path()
        context.stroke()


class TimelineScaleRuler(ScaleRuler, Zoomable):
    """Widget for displaying a ruler which is connected to a timeline.

    Attributes:
        timeline_container (TimelineContainer): The timeline container for
            handling scroll events.
    """

    def __init__(self, timeline_container):
        Zoomable.__init__(self)
        ScaleRuler.__init__(self, self, self.app.settings, self.app.gui.get_style_context())
        self.log("Creating new TimelineScaleRuler")

        self.timeline_container = timeline_container

        timeline_container.timeline.hadj.connect("value-changed", self._hadj_value_changed_cb)

    def _hadj_value_changed_cb(self, hadj):
        """Handles the adjustment value change."""
        self.pixbuf_offset = hadj.get_value()
        self.queue_draw()

# Zoomable interface override

    def zoom_changed(self):
        self.queue_draw()

# Gtk.Widget overrides

    def do_scroll_event(self, event):
        self.timeline_container.timeline.do_scroll_event(event)



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2010, Brandon Lewis <brandon_lewis@berkeley.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
import json
import os.path
from gettext import gettext as _

from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import Gtk

from pitivi.configure import get_audiopresets_dir
from pitivi.configure import get_videopresets_dir
from pitivi.settings import xdg_data_home
from pitivi.utils.loggable import Loggable
from pitivi.utils.ui import alter_style_class


class DeserializeException(Exception):
    pass


class PresetManager(GObject.Object, Loggable):
    """Abstract class for storing a list of presets.

    Subclasses must provide a filename attribute.

    Attributes:
        filename (str): The name of the file where the presets will be stored.
        cur_preset (str): The currently selected preset. Note that a preset
            has to be selected before it can be changed.
        ordered (Gtk.ListStore): A list holding (name, preset_dict) tuples.
        presets (dict): A (name -> preset_dict) map.
        widget_map (dict): A (propname -> (setter_func, getter_func)) map.
            These two functions are used when showing or saving a preset.
    """

    __gsignals__ = {
        "preset-loaded": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, default_path=None, user_path=None, system=None):
        GObject.Object.__init__(self)
        Loggable.__init__(self)

        self.default_path = default_path
        self.user_path = user_path

        self.presets = {}
        self.widget_map = {}
        self.ordered = Gtk.ListStore(str, object)

        def sortme(store, i1, i2, unused_udata):
            name1 = store[i1][0]
            name2 = store[i2][0]
            if name1 > name2:
                return 1

            return -1

        self.ordered.set_sort_func(0, sortme)
        self.ordered.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.cur_preset = None
        # Whether to ignore the update_value calls.
        self.ignore_update_requests = False
        self.system = system

    def setup_ui(self, combo, button):
        # pylint: disable=attribute-defined-outside-init
        self.combo = combo
        self.button = button

        combo.set_model(self.ordered)
        combo.set_id_column(0)
        combo.set_entry_text_column(0)
        combo.connect("changed", self._preset_changed_cb)

        entry = combo.get_child()
        alter_style_class("GtkEntry.unsaved", entry, "font-style:italic;")

        action_group = Gio.SimpleActionGroup()
        menu_model = Gio.Menu()

        action = Gio.SimpleAction.new("new", None)
        action.connect("activate", self._add_preset_cb)
        action_group.add_action(action)
        menu_model.append(_("New"), "preset.%s" % action.get_name())
        self.action_new = action

        action = Gio.SimpleAction.new("remove", None)
        action.connect("activate", self._remove_preset_cb)
        action_group.add_action(action)
        menu_model.append(_("Remove"), "preset.%s" % action.get_name())
        self.action_remove = action

        action = Gio.SimpleAction.new("save", None)
        action.connect("activate", self._save_preset_cb)
        action_group.add_action(action)
        menu_model.append(_("Save"), "preset.%s" % action.get_name())
        self.action_save = action

        menu = Gtk.Menu.new_from_model(menu_model)
        menu.insert_action_group("preset", action_group)
        button.set_popup(menu)

    def _preset_changed_cb(self, combo):
        """Handles the selection of a preset."""
        # Check whether the user selected a preset or editing the preset name.
        self.select_preset(combo)
        self.update_menu_actions()

    def select_preset(self, combo):
        preset_name = combo.get_active_id()
        if preset_name:
            # The user selected a preset.
            self.restore_preset(preset_name)
            self.emit("preset-loaded")

    def _add_preset_cb(self, unused_action, unused_param):
        preset_name = self.get_new_preset_name()
        self.create_preset(preset_name)
        self.combo.set_active_id(preset_name)
        self.update_menu_actions()

    def _remove_preset_cb(self, unused_action, unused_param):
        self.remove_current_preset()
        entry = self.combo.get_child()
        entry.set_text("")
        self.update_menu_actions()

    def _save_preset_cb(self, unused_action, unused_param):
        entry = self.combo.get_child()
        preset_name = entry.get_text()
        self.save_current_preset(preset_name)
        # Useful when a new preset has just been created.
        self.combo.set_active_id(preset_name)
        self.update_menu_actions()

    def update_menu_actions(self):
        entry = self.combo.get_child()
        preset_name = entry.get_text()
        can_save = self.is_save_button_sensitive(preset_name)
        self.action_save.set_enabled(can_save)
        if can_save:
            entry.get_style_context().add_class("unsaved")
        else:
            entry.get_style_context().remove_class("unsaved")

        can_remove = self.is_remove_button_sensitive()
        self.action_remove.set_enabled(can_remove)

        can_create_new = self.is_new_button_sensitive()
        self.action_new.set_enabled(can_create_new)

    def load_all(self):
        self._load_from_dir(self.default_path, extra={"readonly": True})
        if os.path.isfile(self.user_path):
            # We used to save presets as a single file instead of a directory
            os.rename(self.user_path, "%s.old" % self.user_path)
        self._load_from_dir(self.user_path)

    def _load_from_dir(self, presets_dir, extra=None):
        try:
            files = os.listdir(presets_dir)
        except FileNotFoundError:
            self.debug("Presets directory missing: %s", presets_dir)
            return
        for uri in files:
            filepath = os.path.join(presets_dir, uri)
            if filepath.endswith("json"):
                with open(filepath, encoding="UTF-8") as section:
                    parser = json.loads(section.read())
                name = parser["name"]
                if parser.get("removed"):
                    self._forget_preset(name)
                    continue
                try:
                    preset = self._deserialize_preset(parser)
                except DeserializeException as e:
                    self.debug("Failed to load preset %s: %s", filepath, e)
                    continue
                preset["filepath"] = filepath
                if extra:
                    for key, value in extra.items():
                        preset[key] = value
                self._add_preset(name, preset)

    def save_all(self):
        """Writes changes to disk for all presets."""
        for preset_name, unused_values in self.ordered:
            self._save_preset(preset_name)

    def _save_preset(self, preset_name):
        if not os.path.exists(self.user_path):
            os.makedirs(self.user_path)
        try:
            file_path = self.presets[preset_name]["filepath"]
        except KeyError:
            file_path = self._build_file_path(preset_name)
            self.presets[preset_name]["filepath"] = file_path
        with open(file_path, "w", encoding="UTF-8") as fout:
            values = self.presets[preset_name]
            raw = self._serialize_preset(values)
            raw["name"] = preset_name
            serialized = json.dumps(raw, indent=4)
            fout.write(serialized)

    def _build_file_path(self, preset_name):
        file_name = self.system.get_unique_filename(preset_name + ".json")
        return os.path.join(self.user_path, file_name)

    def get_new_preset_name(self):
        """Gets a unique name for a new preset."""
        name = _("New preset")
        i = 1
        while self.has_preset(name):
            name = _("New preset %d") % i
            i += 1
        return name

    def create_preset(self, name, values=None):
        """Creates a preset, overwriting the preset with the same name if any.

        Args:
            name (str): The name of the new preset.
            values (dict): The values of the new preset.
        """
        if not values:
            values = {}
            self._update_preset_values(values)
        self._add_preset(name, values)
        self.cur_preset = name

    def _add_preset(self, name, values):
        self._forget_preset(name)
        self.presets[name] = values
        # Note: This generates a "row-inserted" signal in the model.
        self.ordered.append((name, values))

    def _rename_current_preset(self, new_name):
        """Changes the name of the current preset."""
        old_name = self.cur_preset
        if old_name == new_name:
            # Nothing to do.
            return
        # If there is one already with this name, make way for this one.
        self._forget_preset(new_name)
        for row in self.ordered:
            if row[0] == old_name:
                row[0] = new_name
                break
        self.presets[new_name] = self.presets[old_name]
        if "filepath" in self.presets[old_name]:
            # If the previous preset had already been saved,
            # delete the file and pop it from the list
            self.remove_current_preset()
        else:
            # We're renaming an unsaved preset, so just pop it from the list
            self.presets.pop(old_name)
        new_filepath = self._create_user_preset_path(new_name)
        self.presets[new_name]["filepath"] = new_filepath
        self.cur_preset = new_name

    def _create_user_preset_path(self, preset_name):
        return os.path.join(self.user_path, preset_name + ".json")

    def has_preset(self, name):
        name = name.lower()
        return any(name == preset.lower() for preset in self.get_preset_names())

    def get_preset_names(self):
        return (row[0] for row in self.ordered)

    def update_value(self, name, value):
        """Updates a value in the current preset, if any."""
        if self.ignore_update_requests:
            # This is caused by restore_preset, nothing to do.
            return
        if self.cur_preset:
            self.presets[self.cur_preset][name] = value

    def bind_widget(self, propname, setter_func, getter_func):
        """Links the specified functions to the specified preset property."""
        self.widget_map[propname] = (setter_func, getter_func)

    def restore_preset(self, preset):
        """Selects a preset and copies its values to the widgets.

        Args:
            preset (str): The name of the preset to be selected.
        """
        if preset is None:
            self.cur_preset = None
            return
        if preset not in self.presets:
            return
        self.ignore_update_requests = True
        try:
            values = self.presets[preset]
            self.cur_preset = preset
            for field, (setter, unused_getter) in self.widget_map.items():
                setter(values[field])
        finally:
            self.ignore_update_requests = False

    def save_current_preset(self, new_name=None):
        """Updates the current preset values from the widgets and saves it."""
        if not self.cur_preset:
            self.create_preset(new_name)
        if new_name:
            self._rename_current_preset(new_name)
        values = self.presets[self.cur_preset]
        self._update_preset_values(values)
        self._save_preset(self.cur_preset)

    def _update_preset_values(self, values):
        """Copies the values from the widgets to the specified values dict."""
        for field, (unused_setter, getter) in self.widget_map.items():
            values[field] = getter()

    def _is_current_preset_changed(self, name):
        """Returns whether the widgets values differ from the preset values."""
        if not self.cur_preset:
            # There is no preset selected, nothing to do.
            return False
        if not name == self.cur_preset:
            # The preset can be renamed by saving.
            return True
        values = self.presets[self.cur_preset]
        return any((values[field] != getter()
                    for field, (setter, getter) in self.widget_map.items()))

    def remove_current_preset(self):
        name = self.cur_preset
        preset = self.presets[name]
        filepath = preset.get("filepath")
        if filepath:
            if "readonly" in preset:
                self._mark_removed(name)
            else:
                os.remove(filepath)

        self.cur_preset = None
        self._forget_preset(name)

    def _forget_preset(self, name):
        try:
            self.presets.pop(name)
        except KeyError:
            # Nothing to forget.
            return
        for i, row in enumerate(self.ordered):
            if row[0] == name:
                del self.ordered[i]
                break

    def _mark_removed(self, name):
        data = json.dumps({"name": name, "removed": True}, indent=4)
        filepath = self._create_user_preset_path(name)
        with open(filepath, "w", encoding="UTF-8") as fout:
            fout.write(data)

    def prepend_preset(self, name, values):
        self.presets[name] = values
        # Note: This generates a "row-inserted" signal in the model.
        self.ordered.prepend((name, values))

    def is_save_button_sensitive(self, name):
        """Checks whether the Save button should be enabled.

        Args:
            name (str): The new preset name.
        """
        if self.cur_preset:
            return self._is_current_preset_changed(name)

        if name:
            # Can be saved as new preset.
            return True

        return False

    def is_remove_button_sensitive(self):
        """Checks whether the Remove button should be enabled."""
        if not self.cur_preset:
            return False
        return True

    def is_new_button_sensitive(self):
        """Checks whether the New button should be enabled."""
        return bool(self.cur_preset)

    def _project_to_preset(self, project):
        raise NotImplementedError()

    def matching_preset(self, project):
        query = self._project_to_preset(project)
        for name, preset in self.presets.items():
            matches = True
            for key, value in query.items():
                if not value == preset.get(key):
                    matches = False
                    break
            if matches:
                return name
        return None


class VideoPresetManager(PresetManager):

    def __init__(self, system):
        default_path = get_videopresets_dir()
        user_path = xdg_data_home("video_presets")
        PresetManager.__init__(self, default_path, user_path, system)

    def _deserialize_preset(self, parser):
        width = parser["width"]
        height = parser["height"]

        framerate_num = parser["framerate-num"]
        framerate_denom = parser["framerate-denom"]
        framerate = Gst.Fraction(framerate_num, framerate_denom)

        return {
            "width": width,
            "height": height,
            "frame-rate": framerate,
        }

    def _serialize_preset(self, preset):
        return {
            "width": int(preset["width"]),
            "height": int(preset["height"]),
            "framerate-num": preset["frame-rate"].num,
            "framerate-denom": preset["frame-rate"].denom,
        }

    def _project_to_preset(self, project):
        return {
            "width": project.videowidth,
            "height": project.videoheight,
            "frame-rate": project.videorate}


class AudioPresetManager(PresetManager):

    def __init__(self, system):
        default_path = get_audiopresets_dir()
        user_path = xdg_data_home("audio_presets")
        PresetManager.__init__(self, default_path, user_path, system)

    def _deserialize_preset(self, parser):
        channels = parser["channels"]
        sample_rate = parser["sample-rate"]

        return {
            "channels": channels,
            "sample-rate": sample_rate,
        }

    def _serialize_preset(self, preset):
        return {
            "channels": preset["channels"],
            "sample-rate": int(preset["sample-rate"]),
        }

    def _project_to_preset(self, project):
        return {
            "channels": project.audiochannels,
            "sample-rate": project.audiorate}



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2009, Edward Hervey <bilboed@bilboed.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Rendering-related classes and utilities."""
import os
import posixpath
import time
from enum import IntEnum
from gettext import gettext as _

from gi.repository import GdkPixbuf
from gi.repository import GES
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gst
from gi.repository import GstPbutils
from gi.repository import Gtk

from pitivi import configure
from pitivi.check import MISSING_SOFT_DEPS
from pitivi.dialogs.projectsettings import ProjectSettingsDialog
from pitivi.utils.loggable import Loggable
from pitivi.utils.misc import cmp
from pitivi.utils.misc import is_pathname_valid
from pitivi.utils.misc import path_from_uri
from pitivi.utils.misc import show_user_manual
from pitivi.utils.ripple_update_group import RippleUpdateGroup
from pitivi.utils.ui import AUDIO_CHANNELS
from pitivi.utils.ui import beautify_eta
from pitivi.utils.ui import create_audio_rates_model
from pitivi.utils.ui import create_frame_rates_model
from pitivi.utils.ui import filter_unsupported_media_files
from pitivi.utils.ui import get_combo_value
from pitivi.utils.ui import set_combo_value
from pitivi.utils.widgets import GstElementSettingsDialog


# The category of GstPbutils.EncodingTarget objects holding
# a GstPbutils.EncodingProfile used as a Pitivi render preset.
PITIVI_ENCODING_TARGET_CATEGORY = "user-defined"


def set_icon_and_title(icon, title, preset_item, icon_size=Gtk.IconSize.DND):
    """Adds icon for the respective preset.

    Args:
        icon (Gtk.Image): The image widget to be updated.
        title (Gtk.Label): The label widget to be updated.
        preset_item (PresetItem): Preset profile related information.
        icon_size (Gtk.IconSize): Size of the icon.
    """
    icon_files = {
        "youtube": "youtube.png"
    }
    icon_names = {
        "dvd": "media-optical-dvd-symbolic"
    }

    if not preset_item:
        display_name = _("Custom")
        icon_name = "applications-multimedia-symbolic"
    else:
        display_name = preset_item.display_name
        icon_name = preset_item.name

    title.props.label = display_name
    title.set_xalign(0)
    title.set_yalign(0)

    icon.props.valign = Gtk.Align.START

    if icon_name in icon_files:
        icon_filename = icon_files[icon_name]
        icon_path = os.path.join(configure.get_pixmap_dir(), "presets", icon_filename)

        res, width, height = Gtk.IconSize.lookup(icon_size)
        assert res

        pic = GdkPixbuf.Pixbuf.new_from_file_at_scale(icon_path, height, width, True)
        icon.set_from_pixbuf(pic)
        return

    icon_name = icon_names.get(icon_name, "applications-multimedia-symbolic")
    icon.set_from_icon_name(icon_name, icon_size)


class PresetItem(GObject.Object):
    """Info about a render preset.

    Attributes:
        name (string): Name of the target containing the profile.
        target (GstPbutils.EncodingTarget): The encoding target containing
            the profile.
        profile (GstPbutils.EncodingContainerProfile): The represented preset.
    """

    def __init__(self, name, target, profile):
        GObject.Object.__init__(self)

        name_dict = {
            "youtube": _("YouTube"),
            "dvd": _("DVD"),
        }

        self.name = name
        self.target = target
        self.profile = profile
        self.display_name = name_dict.get(name, name)

    @staticmethod
    def compare_func(item1, item2, *unused_data):
        user_defined1 = item1.target.get_category() == PITIVI_ENCODING_TARGET_CATEGORY
        user_defined2 = item2.target.get_category() == PITIVI_ENCODING_TARGET_CATEGORY
        if user_defined1 != user_defined2:
            return cmp(user_defined1, user_defined2)

        return cmp(item1.name, item2.name)


class PresetBoxRow(Gtk.ListBoxRow):
    """ListBoxRow displaying a render preset.

    Attributes:
        preset_item (PresetItem): Preset profile related information.
    """

    def __init__(self, preset_item):
        Gtk.ListBoxRow.__init__(self)

        self.preset_item = preset_item
        grid = Gtk.Grid()

        title = Gtk.Label()
        icon = Gtk.Image()
        set_icon_and_title(icon, title, preset_item)

        description = Gtk.Label(preset_item.target.get_description())
        description.set_xalign(0)
        description.set_line_wrap(True)
        description.props.max_width_chars = 30

        grid.attach(title, 1, 0, 1, 1)
        grid.attach(description, 1, 1, 1, 1)
        grid.attach(icon, 0, 0, 1, 2)
        grid.set_row_spacing(6)
        grid.set_column_spacing(10)
        grid.set_row_homogeneous(False)
        grid.props.margin = 6
        self.add(grid)


class PresetsManager(GObject.Object, Loggable):
    """Manager of EncodingProfiles used as render presets.

    The EncodingProfiles are retrieved from the available EncodingTargets.
    An EncodingTarget can contain multiple EncodingProfiles.

    The render presets created by us are stored as EncodingProfiles,
    each in its own EncodingTarget.

    Attributes:
        cur_preset_item (PresetItem): The currently selected PresetItem.
        model (Gio.ListStore): The model to store PresetItems for all the preset-profiles.
        project (Project): The project holding the container_profile to be saved or turned into a new preset.
    """

    __gsignals__ = {
        "profile-updated": (GObject.SignalFlags.RUN_LAST, None, (PresetItem,))
    }

    def __init__(self, project):
        GObject.Object.__init__(self)
        Loggable.__init__(self)

        self.project = project

        # menu button actions
        self.action_new = None
        self.action_remove = None
        self.action_save = None

        self.cur_preset_item = None
        self.model = Gio.ListStore.new(PresetItem)

        self.load_all()

    def load_all(self):
        """Loads profiles from GstEncodingTarget and add them to self.model."""
        for target in GstPbutils.encoding_list_all_targets():
            if target.get_category() != GstPbutils.ENCODING_CATEGORY_FILE_EXTENSION:
                self._add_target(target)

    def preset_menubutton_setup(self, button):
        action_group = Gio.SimpleActionGroup()
        menu_model = Gio.Menu()

        action = Gio.SimpleAction.new("new", None)
        action.connect("activate", self._add_preset_cb)
        action_group.add_action(action)
        menu_model.append(_("New"), "preset.%s" % action.get_name())
        self.action_new = action

        action = Gio.SimpleAction.new("remove", None)
        action.connect("activate", self._remove_preset_cb)
        action_group.add_action(action)
        menu_model.append(_("Remove"), "preset.%s" % action.get_name())
        self.action_remove = action

        action = Gio.SimpleAction.new("save", None)
        action.connect("activate", self._save_preset_cb)
        action_group.add_action(action)
        menu_model.append(_("Save"), "preset.%s" % action.get_name())
        self.action_save = action

        self.action_remove.set_enabled(False)
        self.action_save.set_enabled(False)
        menu = Gtk.Menu.new_from_model(menu_model)
        menu.insert_action_group("preset", action_group)
        button.set_popup(menu)

    def _add_preset_cb(self, unused_action, unused_param):
        preset_name = self.get_new_preset_name()
        self.select_preset(self.create_preset(preset_name))

    def _remove_preset_cb(self, unused_action, unused_param):
        self.action_remove.set_enabled(False)
        self.action_save.set_enabled(False)

        if not self.cur_preset_item:
            return

        # There is only one EncodingProfile in the EncodingTarget.
        preset_path = self.cur_preset_item.target.get_path()
        if preset_path:
            os.remove(preset_path)

        res, pos = self.model.find(self.cur_preset_item)
        assert res, self.cur_preset_item.name
        self.model.remove(pos)

        self.cur_preset_item = None
        self.emit("profile-updated", None)

    def _save_preset_cb(self, unused_action, unused_param):
        name = self.cur_preset_item.target.get_name()

        # Remove the currently selected preset item from the model.
        res, pos = self.model.find(self.cur_preset_item)
        assert res, self.cur_preset_item.name
        self.model.remove(pos)

        # Recreate the preset with the current values.
        self.cur_preset_item = self.create_preset(name)
        self.emit("profile-updated", self.cur_preset_item)

    def _add_target(self, encoding_target):
        """Adds the profiles of the specified encoding_target as render presets.

        Args:
            encoding_target (GstPbutils.EncodingTarget): An encoding target.
        """
        preset_items = []
        for profile in encoding_target.get_profiles():
            # The name can be for example "youtube;yt"
            name = encoding_target.get_name().split(";")[0]
            if len(encoding_target.get_profiles()) != 1 and profile.get_name().lower() != "default":
                name += "_" + profile.get_name()

            # Check the GStreamer elements are available.
            profiles = [profile] + profile.get_profiles()
            if not all(self.project.get_element_factory_name(p)
                       for p in profiles):
                self.warning("unusable preset: %s", name)
                continue

            preset_item = PresetItem(name, encoding_target, profile)
            self.model.insert_sorted(preset_item, PresetItem.compare_func)
            preset_items.append(preset_item)

        return preset_items

    def has_preset(self, name):
        name = name.lower()
        preset_names = (item.name for item in self.model)
        return any(name == preset.lower() for preset in preset_names)

    def create_preset(self, preset_name):
        """Creates a preset, overwriting the preset with the same name if any.

        Args:
            preset_name (str): The name for the new preset created.
            values (dict): The values of the new preset.
        """
        target = GstPbutils.EncodingTarget.new(preset_name, PITIVI_ENCODING_TARGET_CATEGORY,
                                               "",
                                               [self.project.container_profile])
        target.save()
        return self._add_target(target)[0]

    def get_new_preset_name(self):
        """Gets a unique name for a new preset."""
        # Translators: This must contain exclusively low case alphanum and '-'
        name = _("new-profile")
        i = 1
        while self.has_preset(name):
            # Translators: This must contain exclusively low case alphanum and '-'
            name = _("new-profile-%d") % i
            i += 1
        return name

    def select_preset(self, preset_item):
        """Selects a preset.

        Args:
            preset_item (PresetItem): The row representing the preset to be applied.
        """
        self.cur_preset_item = preset_item
        writable = bool(preset_item) and \
            len(preset_item.target.get_profiles()) == 1 and \
            os.access(preset_item.target.get_path(), os.W_OK)

        self.action_remove.set_enabled(writable)
        self.action_save.set_enabled(writable)

    def initial_preset(self):
        """Returns the initial preset to be displayed."""
        has_vcodecsettings = bool(self.project.vcodecsettings)
        if has_vcodecsettings:
            # It's a project with previous render settings, see what matches.
            return self.matching_preset()
        else:
            # It's a new project.
            for item in self.model:
                if item.name == "youtube":
                    return item

            return None

    def matching_preset(self):
        """Returns the first preset matching the project's encoding profile."""
        for item in self.model:
            if self.project.matches_container_profile(item.profile):
                return item
        return None


class Encoders(Loggable):
    """Registry of available Muxers, Audio encoders and Video encoders.

    Also keeps the available combinations of those.

    It is a singleton. Use `Encoders()` to access the instance.

    Attributes:
        supported_muxers (List[Gst.ElementFactory]): The supported available
            muxers.
        supported_aencoders (List[Gst.ElementFactory]): The supported available
            audio encoders.
        supported_vencoders (List[Gst.ElementFactory]): The supported available
            video encoders.
        muxers (List[Gst.ElementFactory]): The available muxers.
        aencoders (List[Gst.ElementFactory]): The available audio encoders.
        vencoders (List[Gst.ElementFactory]): The available video encoders.
        compatible_audio_encoders (dict): Maps each muxer name to a list of
            compatible audio encoders ordered by rank.
        compatible_video_encoders (dict): Maps each muxer name to a list of
            compatible video encoders ordered by rank.
        default_muxer (str): The factory name of the default muxer.
        default_audio_encoder (str): The factory name of the default audio
            encoder.
        default_video_encoder (str): The factory name of the default video
            encoder.
    """

    OGG = "oggmux"
    MKV = "matroskamux"
    MP4 = "mp4mux"
    QUICKTIME = "qtmux"
    WEBM = "webmmux"

    if Gst.ElementFactory.find("fdkaacenc"):
        AAC = "fdkaacenc"
    else:
        AAC = "voaacenc"
    AC3 = "avenc_ac3_fixed"
    OPUS = "opusenc"
    VORBIS = "vorbisenc"

    JPEG = "jpegenc"
    THEORA = "theoraenc"
    VP8 = "vp8enc"
    X264 = "x264enc"

    SUPPORTED_ENCODERS_COMBINATIONS = [
        (WEBM, VORBIS, VP8),
        (OGG, VORBIS, THEORA),
        (OGG, OPUS, THEORA),
        (WEBM, OPUS, VP8),
        (MP4, AAC, X264),
        (MP4, AC3, X264),
        (QUICKTIME, AAC, JPEG),
        (MKV, OPUS, X264),
        (MKV, VORBIS, X264),
        (MKV, OPUS, JPEG),
        (MKV, VORBIS, JPEG)]
    """The combinations of muxers and encoders which are supported.

    Mirror of GES_ENCODING_TARGET_COMBINATIONS from
    https://gitlab.freedesktop.org/gstreamer/gst-editing-services/blob/master/tests/validate/geslaunch.py
    """

    _instance = None

    def __new__(cls):
        """Returns the singleton instance."""
        if not cls._instance:
            cls._instance = super(Encoders, cls).__new__(cls)
            # We have to initialize the instance here, otherwise
            # __init__ is called every time we use Encoders().
            Loggable.__init__(cls._instance)
            cls._instance._load_encoders()
            cls._instance._load_combinations()
        return cls._instance

    def _load_encoders(self):
        # pylint: disable=attribute-defined-outside-init
        self.aencoders = []
        self.vencoders = []
        self.muxers = Gst.ElementFactory.list_get_elements(
            Gst.ELEMENT_FACTORY_TYPE_MUXER,
            Gst.Rank.SECONDARY)

        for fact in Gst.ElementFactory.list_get_elements(
                Gst.ELEMENT_FACTORY_TYPE_ENCODER, Gst.Rank.SECONDARY):
            klist = fact.get_klass().split("/")
            if "Video" in klist or "Image" in klist:
                self.vencoders.append(fact)
            elif "Audio" in klist:
                self.aencoders.append(fact)

    def _load_combinations(self):
        # pylint: disable=attribute-defined-outside-init
        self.compatible_audio_encoders = {}
        self.compatible_video_encoders = {}
        useless_muxers = set()
        for muxer in self.muxers:
            aencs = self._find_compatible_encoders(self.aencoders, muxer)
            vencs = self._find_compatible_encoders(self.vencoders, muxer)
            if not aencs or not vencs:
                # The muxer is not compatible with no video encoder or
                # with no audio encoder.
                useless_muxers.add(muxer)
                continue

            muxer_name = muxer.get_name()
            self.compatible_audio_encoders[muxer_name] = aencs
            self.compatible_video_encoders[muxer_name] = vencs

        for muxer in useless_muxers:
            self.muxers.remove(muxer)

        self.factories_by_name = {fact.get_name(): fact
                                  for fact in self.muxers + self.aencoders + self.vencoders}

        good_muxers, good_aencoders, good_vencoders = zip(*self.SUPPORTED_ENCODERS_COMBINATIONS)
        self.supported_muxers = {muxer
                                 for muxer in self.muxers
                                 if muxer.get_name() in good_muxers}
        self.supported_aencoders = {encoder
                                    for encoder in self.aencoders
                                    if encoder.get_name() in good_aencoders}
        self.supported_vencoders = {encoder
                                    for encoder in self.vencoders
                                    if encoder.get_name() in good_vencoders}

        self.default_muxer, \
            self.default_audio_encoder, \
            self.default_video_encoder = self._pick_defaults()

    def _find_compatible_encoders(self, encoders, muxer):
        """Returns the list of encoders compatible with the specified muxer."""
        res = []
        sink_caps = [template.get_caps()
                     for template in muxer.get_static_pad_templates()
                     if template.direction == Gst.PadDirection.SINK]
        for encoder in encoders:
            for template in encoder.get_static_pad_templates():
                if not template.direction == Gst.PadDirection.SRC:
                    continue
                if self._can_muxer_sink_caps(template.get_caps(), sink_caps):
                    res.append(encoder)
                    break
        return sorted(res, key=lambda encoder: - encoder.get_rank())

    def _can_muxer_sink_caps(self, output_caps, sink_caps):
        """Checks whether the specified caps match the muxer's receptors."""
        for caps in sink_caps:
            if not caps.intersect(output_caps).is_empty():
                return True
        return False

    def _pick_defaults(self):
        """Picks the defaults for new projects.

        Returns:
            (str, str, str): The muxer, audio encoder, video encoder.
        """
        for muxer, audio, video in self.SUPPORTED_ENCODERS_COMBINATIONS:
            if muxer not in self.factories_by_name or \
                    audio not in self.factories_by_name or \
                    video not in self.factories_by_name:
                continue
            self.info("Default encoders: %s, %s, %s", muxer, audio, video)
            return muxer, audio, video
        self.warning("No good combination of container and encoders available.")
        return Encoders.OGG, Encoders.VORBIS, Encoders.THEORA

    def is_supported(self, factory):
        """Returns whether the specified factory is supported."""
        if isinstance(factory, str):
            factory = self.factories_by_name[factory]
        return factory in self.supported_muxers or\
            factory in self.supported_aencoders or\
            factory in self.supported_vencoders


def beautify_factory_name(factory):
    """Returns a nice name for the specified Gst.ElementFactory instance.

    Intended for removing redundant words and shorten the codec names.

    Args:
        factory (Gst.ElementFactory): The factory which needs to be displayed.

    Returns:
        str: Cleaned up name.
    """
    # Only replace lowercase versions of "format", "video", "audio"
    # otherwise they might be part of a trademark name.
    words_to_remove = ["Muxer", "muxer", "Encoder", "encoder",
                       "format", "video", "audio", "instead",
                       # Incorrect naming for Sorenson Spark:
                       "Flash Video (FLV) /", ]
    words_to_replace = [["version ", "v"], ["Microsoft", "MS"], ]
    name = factory.get_longname()
    for word in words_to_remove:
        name = name.replace(word, "")
    for match, replacement in words_to_replace:
        name = name.replace(match, replacement)
    return " ".join(word for word in name.split())


def extension_for_muxer(muxer_name):
    """Returns the file extension appropriate for the specified muxer.

    Args:
        muxer_name (str): The name of the muxer factory.
    """
    exts = {
        "asfmux": "asf",
        "avimux": "avi",
        "avmux_3g2": "3g2",
        "avmux_avm2": "avm2",
        "avmux_dvd": "vob",
        "avmux_flv": "flv",
        "avmux_ipod": "mp4",
        "avmux_mpeg": "mpeg",
        "avmux_mpegts": "mpeg",
        "avmux_psp": "mp4",
        "avmux_rm": "rm",
        "avmux_svcd": "mpeg",
        "avmux_swf": "swf",
        "avmux_vcd": "mpeg",
        "avmux_vob": "vob",
        "flvmux": "flv",
        "gppmux": "3gp",
        "matroskamux": "mkv",
        "mj2mux": "mj2",
        "mp4mux": "mp4",
        "mpegpsmux": "mpeg",
        "mpegtsmux": "mpeg",
        "mvemux": "mve",
        "mxfmux": "mxf",
        "oggmux": "ogv",
        "qtmux": "mov",
        "webmmux": "webm"}
    return exts.get(muxer_name)


class Quality(IntEnum):
    LOW = 0
    MEDIUM = 1
    HIGH = 2


class QualityAdapter(Loggable):
    """Adapter between a quality value and the properties of an Encoder."""

    def __init__(self, props_values, prop_name=None):
        super().__init__()

        self.props_values = props_values

        if not prop_name:
            assert len(props_values) == 1
            prop_name = list(props_values.keys())[0]
        self.prop_name = prop_name

    def calculate_quality(self, vcodecsettings):
        if self.prop_name in vcodecsettings:
            encoder_property_value = vcodecsettings[self.prop_name]
            values = self.props_values[self.prop_name]
            for quality in (Quality.HIGH, Quality.MEDIUM, Quality.LOW):
                if (values[0] < values[-1] and encoder_property_value >= values[quality]) or \
                        (values[0] > values[-1] and encoder_property_value <= values[quality]):
                    break

            self.debug("Got existing value for prop %s=%s -> quality=%s", self.prop_name, encoder_property_value, quality)
        else:
            quality = Quality.LOW
            self.debug("Cannot calculate quality from missing prop %s", self.prop_name)

        return quality

    def update_project_vcodecsettings(self, project, quality):
        for prop_name, values in self.props_values.items():
            if callable(values):
                value = values(project)
            else:
                value = values[quality]
            project.vcodecsettings[prop_name] = value


quality_adapters = {
    Encoders.X264: QualityAdapter(
        {
            # quantizer accepts values between 0..50, default is 21.
            # Values inspired by https://slhck.info/video/2017/03/01/rate-control.html
            "quantizer": (25, 21, 18),
            # Encoding pass/type: Constant Quality
            # https://gstreamer.freedesktop.org/documentation/x264/index.html?gi-language=python#GstX264EncPass
            "pass": lambda unused_project: 5,
        },
        prop_name="quantizer"),
    Encoders.VP8: QualityAdapter(
        {
            # cq-level accepts values between 0..63, default is 10.
            "cq-level": (31, 47, 63),
            # Rate control mode: Constant Quality Mode (CQ) mode
            # https://gstreamer.freedesktop.org/documentation/vpx/GstVPXEnc.html?gi-language=python#GstVPXEnc:end-usage
            "end-usage": lambda unused_project: 2,
        },
        prop_name="cq-level"),
    Encoders.THEORA: QualityAdapter({
        # Setting the quality property will produce a variable bitrate (VBR) stream.
        # quality accepts values between 0..63, default is 48.
        "quality": (31, 48, 63)}),
    Encoders.JPEG: QualityAdapter({
        # quality accepts values between 0..100, default is 85.
        "quality": (70, 85, 100)}),
}


class RenderingProgressDialog(GObject.Object):

    __gsignals__ = {
        "pause": (GObject.SignalFlags.RUN_LAST, None, ()),
        "cancel": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, app, parent):
        GObject.Object.__init__(self)

        self.app = app
        self.main_render_dialog = parent
        self.builder = Gtk.Builder()
        self.builder.add_from_file(
            os.path.join(configure.get_ui_dir(), "renderingprogress.ui"))
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("render-progress")
        self.table1 = self.builder.get_object("table1")
        self.progressbar = self.builder.get_object("progressbar")
        self.play_pause_button = self.builder.get_object("play_pause_button")
        self.play_rendered_file_button = self.builder.get_object(
            "play_rendered_file_button")
        self.close_button = self.builder.get_object("close_button")
        self.show_in_file_manager_button = self.builder.get_object(
            "show_in_file_manager_button")
        self.cancel_button = self.builder.get_object("cancel_button")
        self._filesize_est_label = self.builder.get_object(
            "estimated_filesize_label")
        self._filesize_est_value_label = self.builder.get_object(
            "estimated_filesize_value_label")
        # Parent the dialog with mainwindow, since renderingdialog is hidden.
        # It allows this dialog to properly minimize together with mainwindow
        self.window.set_transient_for(self.app.gui)
        self.window.set_icon_name("system-run-symbolic")

        self.play_rendered_file_button.get_style_context().add_class("suggested-action")

        # We will only show the close/play buttons when the render is done:
        self.play_rendered_file_button.hide()
        self.close_button.hide()
        self.show_in_file_manager_button.hide()

    def update_position(self, fraction):
        self.progressbar.set_fraction(fraction)
        self.window.set_title(
            _("Rendering — %d%% complete") % int(100 * fraction))

    def update_progressbar_eta(self, time_estimation):
        # Translators: this string indicates the estimated time
        # remaining until an action (such as rendering) completes.
        # The "%s" is an already-localized human-readable duration,
        # such as "31 seconds", "1 minute" or "1 hours, 14 minutes".
        # In some languages, "About %s left" can be expressed roughly as
        # "There remains approximatively %s" (to handle gender and plurals).
        self.progressbar.set_text(_("About %s left") % time_estimation)

    def set_filesize_estimate(self, estimated_filesize=None):
        if not estimated_filesize:
            self._filesize_est_label.hide()
            self._filesize_est_value_label.hide()
        else:
            self._filesize_est_value_label.set_text(estimated_filesize)
            self._filesize_est_label.show()
            self._filesize_est_value_label.show()

    def _delete_event_cb(self, unused_dialog_widget, unused_event):
        """Stops the rendering."""
        # The user closed the window by pressing Escape.
        self.emit("cancel")

    def _cancel_button_clicked_cb(self, unused_button):
        self.emit("cancel")

    def _pause_button_clicked_cb(self, unused_button):
        self.emit("pause")

    def _close_button_clicked_cb(self, unused_button):
        self.window.destroy()
        if self.main_render_dialog.notification is not None:
            self.main_render_dialog.notification.close()
        self.main_render_dialog.window.show()

    def _play_rendered_file_button_clicked_cb(self, unused_button):
        Gio.AppInfo.launch_default_for_uri(self.main_render_dialog.outfile, None)

    def _show_in_file_manager_button_clicked_cb(self, unused_button):
        directory_uri = posixpath.dirname(self.main_render_dialog.outfile)
        Gio.AppInfo.launch_default_for_uri(directory_uri, None)


class RenderDialog(Loggable):
    """Render dialog box.

    Args:
        app (Pitivi): The app.
        project (Project): The project to be rendered.

    Attributes:
        preferred_aencoder (str): The last audio encoder selected by the user.
        preferred_vencoder (str): The last video encoder selected by the user.
    """

    INHIBIT_REASON = _("Currently rendering")

    _factory_formats = {}

    def __init__(self, app, project):
        Loggable.__init__(self)

        self.app = app
        self.project = project
        self._pipeline = self.project.pipeline

        self.outfile = None
        self.notification = None

        # Variables to keep track of progress indication timers:
        self._filesize_estimate_timer = None
        self._time_estimate_timer = None
        self._is_rendering = False
        self._rendering_is_paused = False
        self._last_timestamp_when_pausing = 0
        self.current_position = None
        self._time_started = 0
        self._time_spent_paused = 0  # Avoids the ETA being wrong on resume
        self._is_filename_valid = True

        # Various gstreamer signal connection ID's
        # {object: sigId}
        self._gst_signal_handlers_ids = {}

        self.presets_manager = PresetsManager(project)

        # Whether encoders changing are a result of changing the muxer.
        self.muxer_combo_changing = False
        self.progress = None
        self.dialog = None
        self.preset_listbox = None
        self._create_ui()

        # Directory and Filename
        if not self.project.name:
            self._update_filename(_("Untitled"))
        else:
            self._update_filename(self.project.name)

        self._setting_encoding_profile = False

        # We store these so that when the user tries various container formats,
        # (AKA muxers) we select these a/v encoders, if they are compatible with
        # the current container format.
        self.preferred_vencoder = self.project.vencoder
        self.preferred_aencoder = self.project.aencoder
        self.__replaced_assets = {}

        self._display_render_settings()

        self.window.connect("delete-event", self._delete_event_cb)
        self.project.connect("video-size-changed", self._project_video_size_changed_cb)

        self.presets_manager.connect("profile-updated", self._presets_manager_profile_updated_cb)

        preset_item: PresetItem = self.presets_manager.initial_preset()
        if preset_item:
            if self.apply_preset(preset_item):
                self.apply_vcodecsettings_quality(Quality.MEDIUM)
                self.presets_manager.select_preset(preset_item)

        set_icon_and_title(self.preset_icon, self.preset_label, self.presets_manager.cur_preset_item)
        self._update_quality_scale()

        # Monitor changes to keep the preset_selection_menubutton updated.
        self.widgets_group = RippleUpdateGroup()
        self.widgets_group.add_vertex(self.preset_selection_menubutton,
                                      update_func=self._update_preset_selection_menubutton_func)

        self.widgets_group.add_vertex(self.muxer_combo, signal="changed")
        self.widgets_group.add_vertex(self.video_encoder_combo, signal="changed")
        self.widgets_group.add_vertex(self.frame_rate_combo, signal="changed")
        self.widgets_group.add_vertex(self.audio_encoder_combo, signal="changed")
        self.widgets_group.add_vertex(self.channels_combo, signal="changed")
        self.widgets_group.add_vertex(self.sample_rate_combo, signal="changed")

        self.widgets_group.add_edge(self.muxer_combo, self.preset_selection_menubutton)
        self.widgets_group.add_edge(self.video_encoder_combo, self.preset_selection_menubutton)
        self.widgets_group.add_edge(self.frame_rate_combo, self.preset_selection_menubutton)
        self.widgets_group.add_edge(self.audio_encoder_combo, self.preset_selection_menubutton)
        self.widgets_group.add_edge(self.channels_combo, self.preset_selection_menubutton)
        self.widgets_group.add_edge(self.sample_rate_combo, self.preset_selection_menubutton)

    def _presets_manager_profile_updated_cb(self, presets_manager, preset_item):
        """Handles the saving or removing of a render preset."""
        set_icon_and_title(self.preset_icon, self.preset_label, preset_item)

    def _set_encoding_profile(self, encoding_profile):
        """Sets the encoding profile of the project.

        Args:
            encoding_profile(GstPbutils.EncodingContainerProfile): The profile to set.

        Returns:
            bool: Whether the operation succeeded.
        """
        self.project.set_container_profile(encoding_profile)
        self._setting_encoding_profile = True
        try:
            muxer = Encoders().factories_by_name.get(self.project.muxer)
            if not set_combo_value(self.muxer_combo, muxer):
                self.error("Failed to set muxer_combo to %s", muxer)
                return False

            self.update_available_encoders()
            self._update_valid_audio_restrictions(Gst.ElementFactory.find(self.project.aencoder))
            self._update_valid_video_restrictions(Gst.ElementFactory.find(self.project.vencoder))
            aencoder = Encoders().factories_by_name.get(self.project.aencoder)
            vencoder = Encoders().factories_by_name.get(self.project.vencoder)
            for i, (combo, name, value) in enumerate([
                    (self.audio_encoder_combo, "aencoder", aencoder),
                    (self.video_encoder_combo, "vencoder", vencoder),
                    (self.sample_rate_combo, "audiorate", self.project.audiorate),
                    (self.channels_combo, "audiochannels", self.project.audiochannels),
                    (self.frame_rate_combo, "videorate", self.project.videorate)]):
                if value is None:
                    self.error("%d - Got no value for %s (%s)... rolling back",
                               i, name, combo)
                    return False

                if not set_combo_value(combo, value):
                    self.error("%d - Could not set value %s for combo %s... rolling back",
                               i, value, combo)
                    return False

            self.update_resolution()
            self.project.add_encoding_profile(self.project.container_profile)
            self._update_file_extension()
        finally:
            self._setting_encoding_profile = False

        return True

    def _create_ui(self):
        builder = Gtk.Builder()
        builder.add_from_file(
            os.path.join(configure.get_ui_dir(), "renderingdialog.ui"))
        builder.connect_signals(self)

        self.window = builder.get_object("render-dialog")
        self.video_output_checkbutton = builder.get_object(
            "video_output_checkbutton")
        self.audio_output_checkbutton = builder.get_object(
            "audio_output_checkbutton")
        self.render_button = builder.get_object("render_button")
        self.video_settings_button = builder.get_object(
            "video_settings_button")
        self.audio_settings_button = builder.get_object(
            "audio_settings_button")
        self.frame_rate_combo = builder.get_object("frame_rate_combo")
        self.frame_rate_combo.set_model(create_frame_rates_model())
        self.scale_spinbutton = builder.get_object("scale_spinbutton")
        self.channels_combo = builder.get_object("channels_combo")
        self.channels_combo.set_model(AUDIO_CHANNELS)
        self.sample_rate_combo = builder.get_object("sample_rate_combo")
        self.muxer_combo = builder.get_object("muxercombobox")
        self.audio_encoder_combo = builder.get_object("audio_encoder_combo")
        self.video_encoder_combo = builder.get_object("video_encoder_combo")
        self.filebutton = builder.get_object("filebutton")
        self.fileentry = builder.get_object("fileentry")
        self.resolution_label = builder.get_object("resolution_label")
        self.preset_selection_menubutton = builder.get_object("preset_selection_menubutton")
        self.preset_label = builder.get_object("preset_label")
        self.preset_icon = builder.get_object("preset_icon")
        self.preset_menubutton = builder.get_object("preset_menubutton")
        self.preset_popover = builder.get_object("preset_popover")
        self.quality_box = builder.get_object("quality_box")
        self.quality_scale = builder.get_object("quality_scale")
        self.quality_adjustment = self.quality_scale.props.adjustment

        self.quality_adjustment_handler_id = self.quality_adjustment.connect("value-changed", self._quality_adjustment_value_changed_cb)

        # round_digits is set to -1 in gtk_scale_set_draw_value.
        # Set it to 0 since we don't care about intermediary values.
        self.quality_scale.props.round_digits = 0

        lower = self.quality_adjustment.props.lower
        upper = self.quality_adjustment.props.upper
        self.quality_scale.add_mark(lower + (upper - lower) / 2, Gtk.PositionType.BOTTOM, _("medium"))
        self.quality_scale.add_mark(upper, Gtk.PositionType.BOTTOM, _("high"))

        self.__automatically_use_proxies = builder.get_object(
            "automatically_use_proxies")

        set_icon_and_title(self.preset_icon, self.preset_label, None)
        self.preset_selection_menubutton.connect("clicked", self._preset_selection_menubutton_clicked_cb)

        self.__always_use_proxies = builder.get_object("always_use_proxies")
        self.__always_use_proxies.props.group = self.__automatically_use_proxies

        self.__never_use_proxies = builder.get_object("never_use_proxies")
        self.__never_use_proxies.props.group = self.__automatically_use_proxies

        self.presets_manager.preset_menubutton_setup(self.preset_menubutton)

        self.window.set_icon_name("system-run-symbolic")
        self.window.set_transient_for(self.app.gui)

        media_types = self.project.ges_timeline.ui.media_types

        self.audio_output_checkbutton.props.active = media_types & GES.TrackType.AUDIO
        self._update_audio_widgets_sensitivity()

        self.video_output_checkbutton.props.active = media_types & GES.TrackType.VIDEO
        self._update_video_widgets_sensitivity()

        self.listbox_setup()

    def listbox_setup(self):
        self.preset_listbox = Gtk.ListBox()
        self.preset_listbox.set_selection_mode(Gtk.SelectionMode.NONE)

        self.preset_listbox.bind_model(self.presets_manager.model, self._create_preset_row_func)

        self.preset_listbox.connect("row-activated", self._preset_listbox_row_activated_cb)
        self.preset_popover.add(self.preset_listbox)

    def _create_preset_row_func(self, preset_item):
        return PresetBoxRow(preset_item)

    def _preset_listbox_row_activated_cb(self, listbox, row):
        if self.apply_preset(row.preset_item):
            quality = Quality.MEDIUM
            if self.quality_scale.get_sensitive():
                quality = self.quality_adjustment.props.value

            self.apply_vcodecsettings_quality(quality)
            self._update_quality_scale()

            self.preset_popover.hide()

    def apply_preset(self, preset_item: PresetItem):
        old_profile = self.project.container_profile
        profile = preset_item.profile.copy()
        if not self._set_encoding_profile(profile):
            self.error("failed to apply the encoding profile, reverting to previous one")
            self._set_encoding_profile(old_profile)
            return False

        self.presets_manager.select_preset(preset_item)
        set_icon_and_title(self.preset_icon, self.preset_label, preset_item)

        return True

    def _preset_selection_menubutton_clicked_cb(self, button):
        self.preset_popover.show_all()

    def _project_video_size_changed_cb(self, project):
        """Handles Project metadata changes."""
        self.update_resolution()

    def create_combobox_model(self, factories):
        """Creates a model for a combobox showing factories.

        Args:
            combobox (Gtk.ComboBox): The combobox to setup.
            factories (List[Gst.ElementFactory]): The factories to display.

        Returns:
            Gtk.ListStore: The model with (display name, factory, unsupported).
        """
        model = Gtk.TreeStore(str, object)
        data_supported = []
        data_unsupported = []
        for factory in factories:
            supported = Encoders().is_supported(factory)
            row = (beautify_factory_name(factory), factory)
            if supported:
                data_supported.append(row)
            else:
                data_unsupported.append(row)

        data_supported.sort()
        for row in data_supported:
            model.append(None, row)

        if data_unsupported:
            # Translators: This item appears in a combobox's popup and
            # contains as children the unsupported (but still available)
            # muxers and encoders.
            unsupported_iter = model.append(None, (_("Unsupported"), None))
            data_unsupported.sort()
            for row in data_unsupported:
                model.append(unsupported_iter, row)

        return model

    def _display_settings(self):
        """Applies the project settings to the UI."""
        # Video settings
        fr_datum = (self.project.videorate.num, self.project.videorate.denom)
        model = create_frame_rates_model(fr_datum)
        self.frame_rate_combo.set_model(model)
        set_combo_value(self.frame_rate_combo, self.project.videorate)

        # Audio settings
        res = set_combo_value(self.channels_combo, self.project.audiochannels)
        assert res, self.project.audiochannels
        res = set_combo_value(self.sample_rate_combo, self.project.audiorate)
        assert res, self.project.audiorate

    def _update_audio_widgets_sensitivity(self):
        active = self.audio_output_checkbutton.get_active()
        self.channels_combo.set_sensitive(active)
        self.sample_rate_combo.set_sensitive(active)
        self.audio_encoder_combo.set_sensitive(active)
        self.audio_settings_button.set_sensitive(active)
        self.project.audio_profile.set_enabled(active)
        self.__update_render_button_sensitivity()

    def _update_video_widgets_sensitivity(self):
        active = self.video_output_checkbutton.get_active()
        self.scale_spinbutton.set_sensitive(active)
        self.frame_rate_combo.set_sensitive(active)
        self.video_encoder_combo.set_sensitive(active)
        self.video_settings_button.set_sensitive(active)
        self.project.video_profile.set_enabled(active)
        self.__update_render_button_sensitivity()

    def _display_render_settings(self):
        """Applies the project render settings to the UI."""
        # Video settings
        # This will trigger an update of the video resolution label.
        self.scale_spinbutton.set_value(self.project.render_scale)

        # Muxer settings
        model = self.create_combobox_model(Encoders().muxers)
        self.muxer_combo.set_model(model)
        # This will trigger an update of the codec comboboxes.
        muxer = Encoders().factories_by_name.get(self.project.muxer)
        if muxer:
            if not set_combo_value(self.muxer_combo, muxer):
                # The project's muxer is not available on this system.
                # Pick the first one available.
                first = self.muxer_combo.props.model.get_iter_first()
                set_combo_value(self.muxer_combo, first)

    def _check_filename(self):
        """Displays a warning if the file path already exists."""
        filepath = self.fileentry.get_text()
        if not filepath:
            tooltip_text = _("A file name is required.")
            self._is_filename_valid = False
        else:
            filepath = os.path.realpath(filepath)
            if os.path.isdir(filepath):
                tooltip_text = _("A file name is required.")
                self._is_filename_valid = False
            elif os.path.exists(filepath):
                tooltip_text = _("This file already exists.\n"
                                 "If you don't want to overwrite it, choose a "
                                 "different file name or folder.")
                self._is_filename_valid = True
            elif not is_pathname_valid(filepath):
                tooltip_text = _("Invalid file path")
                self._is_filename_valid = False
            else:
                tooltip_text = None
                self._is_filename_valid = True

        warning_icon = "dialog-warning-symbolic" if tooltip_text else None
        self.fileentry.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, warning_icon)
        self.fileentry.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, tooltip_text)
        self.__update_render_button_sensitivity()

    def _get_filesize_estimate(self):
        """Estimates the final file size.

        Estimates in megabytes (over 30 MB) are rounded to the nearest 10 MB
        to smooth out small variations. You'd be surprised how imprecision can
        improve perceived accuracy.

        Returns:
            str: A human-readable (ex: "14 MB") estimate for the file size.
        """
        if not self.current_position:
            return None

        current_filesize = os.stat(path_from_uri(self.outfile)).st_size
        length = self.project.ges_timeline.props.duration
        estimated_size = current_filesize * length / self.current_position
        # Now let's make it human-readable (instead of octets).
        # If it's in the giga range (10⁹) instead of mega (10⁶), use 2 decimals
        gigabytes = estimated_size / (10 ** 9)
        if gigabytes >= 1:
            return _("%.2f GB") % gigabytes

        megabytes = int(estimated_size / (10 ** 6))
        if megabytes == 0:
            return None
        elif megabytes > 30:
            megabytes = int(round(megabytes, -1))  # -1 means round to 10
        return _("%d MB") % megabytes

    def _update_filename(self, basename):
        """Updates the filename UI element to show the specified file name."""
        extension = extension_for_muxer(self.project.muxer)
        if extension:
            name = "%s%s%s" % (basename, os.path.extsep, extension)
        else:
            name = basename

        self.fileentry.set_text(os.path.join(self.app.settings.lastExportFolder, name))

    def _update_valid_restriction_values(self, caps, combo, caps_template,
                                         model, combo_value,
                                         caps_template_expander=None):
        def caps_template_expander_func(caps_template, value):
            return caps_template % value

        if not caps_template_expander:
            caps_template_expander = caps_template_expander_func

        model_headers = [model.get_column_type(i) for i in range(model.get_n_columns())]
        reduced_model = Gtk.ListStore(*model_headers)
        reduced = []
        for name, value in dict(model).items():
            caps_raw = caps_template_expander(caps_template, value)
            ecaps = Gst.Caps(caps_raw)
            if caps.intersect(ecaps).is_empty():
                self.warning(
                    "Ignoring value because not supported by the encoder: %s",
                    caps_raw)
            else:
                reduced.append((name, value))

        for value in sorted(reduced, key=lambda v: float(v[1])):
            reduced_model.append(value)
        combo.set_model(reduced_model)

        if not set_combo_value(combo, combo_value):
            combo.set_active(len(reduced_model) - 1)
            self.warning("%s in %s not supported, setting: %s",
                         combo_value, caps_template, get_combo_value(combo))

    def _update_valid_audio_restrictions(self, factory):
        template = [t for t in factory.get_static_pad_templates()
                    if t.direction == Gst.PadDirection.SINK][0]

        caps = template.static_caps.get()
        model = create_audio_rates_model(self.project.audiorate)
        self._update_valid_restriction_values(caps, self.sample_rate_combo,
                                              "audio/x-raw,rate=(int)%d",
                                              model,
                                              self.project.audiorate)

        self._update_valid_restriction_values(caps, self.channels_combo,
                                              "audio/x-raw,channels=(int)%d",
                                              AUDIO_CHANNELS,
                                              self.project.audiochannels)

    def _update_valid_video_restrictions(self, factory):
        def fraction_expander_func(caps_template, value):
            return caps_template % (value.num, value.denom)

        template = [t for t in factory.get_static_pad_templates()
                    if t.direction == Gst.PadDirection.SINK][0]

        caps = template.static_caps.get()

        fr_datum = (self.project.videorate.num, self.project.videorate.denom)
        model = create_frame_rates_model(fr_datum)
        self._update_valid_restriction_values(
            caps, self.frame_rate_combo,
            "video/x-raw,framerate=(GstFraction)%d/%d", model,
            self.project.videorate,
            caps_template_expander=fraction_expander_func)

    def update_available_encoders(self):
        """Updates the encoder comboboxes to show the available encoders."""
        self.muxer_combo_changing = True
        try:
            model = self.create_combobox_model(
                Encoders().compatible_video_encoders[self.project.muxer])
            self.video_encoder_combo.set_model(model)
            self._update_encoder_combo(self.video_encoder_combo,
                                       self.preferred_vencoder)

            model = self.create_combobox_model(
                Encoders().compatible_audio_encoders[self.project.muxer])
            self.audio_encoder_combo.set_model(model)
            self._update_encoder_combo(self.audio_encoder_combo,
                                       self.preferred_aencoder)
        finally:
            self.muxer_combo_changing = False

    def _update_encoder_combo(self, encoder_combo, preferred_encoder):
        """Selects the specified encoder for the specified encoder combo."""
        if preferred_encoder:
            encoder = Encoders().factories_by_name.get(preferred_encoder)
            if set_combo_value(encoder_combo, encoder):
                # The preference was found in the combo's model
                # and has been activated.
                return

        # Pick the first encoder from the combobox's model.
        first = encoder_combo.props.model.get_iter_first()
        if not first:
            # Model is empty. Should not happen.
            self.warning("Model is empty")
            return

        if encoder_combo.props.model.iter_has_child(first):
            # There are no supported encoders and the first element is
            # the Unsupported group. Activate its first child.
            first = encoder_combo.props.model.iter_nth_child(first, 0)

        encoder_combo.set_active_iter(first)

    def _element_settings_dialog(self, factory, media_type):
        """Opens a dialog to edit the properties for the specified factory.

        Args:
            factory (Gst.ElementFactory): The factory for editing.
            media_type (str): String describing the media type ("audio" or "video")
        """
        # Reconstitute the property name from the media type (vcodecsettings or acodecsettings)
        properties = getattr(self.project, media_type[0] + "codecsettings")

        self.dialog = GstElementSettingsDialog(factory, properties=properties,
                                               caps=getattr(self.project, media_type + "_profile").get_format(),
                                               parent_window=self.window)
        self.dialog.ok_btn.connect(
            "clicked", self._ok_button_clicked_cb, media_type)

    def __additional_debug_info(self):
        if self.project.vencoder == "x264enc":
            if self.project.videowidth % 2 or self.project.videoheight % 2:
                return "\n\n%s\n\n" % _("<b>Make sure your rendering size is even, "
                                        "x264enc might not be able to render otherwise.</b>\n\n")

        return ""

    def _show_render_error_dialog(self, error, unused_details):
        primary_message = _("Sorry, something didn’t work right.")
        secondary_message = "".join([
            _("An error occurred while trying to render your project."),
            self.__additional_debug_info(),
            _("You might want to check our troubleshooting guide or file a bug report. "
              "The GStreamer error was:"),
            "\n\n<i>" + str(error) + "</i>"])

        dialog = Gtk.MessageDialog(transient_for=self.window, modal=True,
                                   message_type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK,
                                   text=primary_message)
        dialog.set_property("secondary-text", secondary_message)
        dialog.set_property("secondary-use-markup", True)
        dialog.show_all()
        dialog.run()
        dialog.destroy()

    def start_action(self):
        """Starts the render process."""
        self._pipeline.set_state(Gst.State.NULL)
        self._pipeline.set_mode(GES.PipelineFlags.RENDER)
        encodebin = self._pipeline.get_by_name("internal-encodebin")
        self._gst_signal_handlers_ids[encodebin] = encodebin.connect(
            "element-added", self.__element_added_cb)
        for element in encodebin.iterate_recurse():
            self.__set_properties(element)
        self._pipeline.set_state(Gst.State.PLAYING)
        self._is_rendering = True
        self._time_started = time.time()

    def _cancel_render(self, *unused_args):
        self.debug("Aborting render")
        self._shut_down()
        self._destroy_progress_window()

    def _shut_down(self):
        """Shuts down the pipeline and disconnects from its signals."""
        self._is_rendering = False
        self._rendering_is_paused = False
        self._time_spent_paused = 0
        self._pipeline.set_state(Gst.State.NULL)
        self.project.set_rendering(False)
        self._use_proxy_assets()
        self._disconnect_from_gst()
        self._pipeline.set_mode(GES.PipelineFlags.FULL_PREVIEW)
        self._pipeline.set_state(Gst.State.PAUSED)

    def _pause_render(self, unused_progress):
        self._rendering_is_paused = self.progress.play_pause_button.get_active()
        if self._rendering_is_paused:
            self._last_timestamp_when_pausing = time.time()
        else:
            self._time_spent_paused += time.time() - self._last_timestamp_when_pausing
            self.debug(
                "Resuming render after %d seconds in pause", self._time_spent_paused)
        self.project.pipeline.toggle_playback()

    def _destroy_progress_window(self):
        """Handles the completion or the cancellation of the render process."""
        self.progress.window.destroy()
        self.progress = None
        self.window.show()  # Show the rendering dialog again

    def _disconnect_from_gst(self):
        for obj, handler_id in self._gst_signal_handlers_ids.items():
            obj.disconnect(handler_id)
        self._gst_signal_handlers_ids = {}
        try:
            self.project.pipeline.disconnect_by_func(self._update_position_cb)
        except TypeError:
            # The render was successful, so this was already disconnected
            pass

    def destroy(self):
        self.window.destroy()

    def _maybe_play_finished_sound(self):
        """Plays a sound to signal the render operation is done."""
        if "GSound" in MISSING_SOFT_DEPS:
            return

        from gi.repository import GSound
        sound_context = GSound.Context()
        try:
            sound_context.init()
            sound_context.play_simple({GSound.ATTR_EVENT_ID: "complete"})
        except GLib.Error as e:
            self.warning("GSound failed to play: %s", e)

    def __unset_effect_preview_props(self):
        for clip in self.project.ges_timeline.ui.clips():
            for effect in clip.get_top_effects():
                effect_name = effect.get_property("bin-description")
                if effect_name == "frei0r-filter-3-point-color-balance":
                    effect.set_child_property("split-preview", False)

    def _asset_replacement(self, clip):
        if not isinstance(clip, GES.UriClip):
            return None

        asset = clip.get_asset()
        asset_target = asset.get_proxy_target()
        if not asset_target:
            # The asset is not a proxy.
            return None

        # Replace all proxies
        if self.__never_use_proxies.get_active():
            return asset_target

        # Use HQ Proxy (or equivalent) only for unsupported assets
        if self.__automatically_use_proxies.get_active():
            if self.app.proxy_manager.is_asset_format_well_supported(asset_target):
                return asset_target

            proxy_unsupported = True

        # Use HQ Proxy (or equivalent) whenever available
        if self.__always_use_proxies.get_active() or proxy_unsupported:
            if self.app.proxy_manager.is_hq_proxy(asset):
                return None

            if self.app.proxy_manager.is_scaled_proxy(asset):
                width, height = self.project.get_video_width_and_height(render=True)
                stream = asset.get_info().get_video_streams()[0]
                asset_res = [stream.get_width(), stream.get_height()]

                if asset_res[0] == width and asset_res[1] == height:
                    # Check whether the scaled proxy size matches the render size
                    # exactly. If the size is same, render from the scaled proxy
                    # to avoid double scaling.
                    return None

                hq_proxy = GES.Asset.request(GES.UriClip,
                                             self.app.proxy_manager.get_proxy_uri(asset_target))
                return hq_proxy
        return None

    def __replace_proxies(self):
        for clip in self.project.ges_timeline.ui.clips():
            asset = self._asset_replacement(clip)
            if asset:
                self.__replaced_assets[clip] = clip.get_asset()
                clip.set_asset(asset)

    def _use_proxy_assets(self):
        for clip, asset in self.__replaced_assets.items():
            self.info("Reverting to using proxy asset %s", asset)
            clip.set_asset(asset)

        self.__replaced_assets = {}

    # ------------------- Callbacks ------------------------------------------ #

    def _ok_button_clicked_cb(self, unused_button, media_type):
        assert media_type in ("audio", "video")
        setattr(self.project, media_type[0] + "codecsettings", self.dialog.get_settings())

        caps = self.dialog.get_caps()
        if caps:
            getattr(self.project, media_type + "_profile").set_format(caps)
        self.dialog.window.destroy()

    def _select_file_clicked_cb(self, unused_button):
        chooser = Gtk.FileChooserNative.new(
            _("Select file path to render"),
            self.window,
            Gtk.FileChooserAction.SAVE,
            None, None)

        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("Supported file formats"))
        file_filter.add_custom(Gtk.FileFilterFlags.URI | Gtk.FileFilterFlags.MIME_TYPE,
                               filter_unsupported_media_files)
        chooser.add_filter(file_filter)
        chooser.set_current_folder(self.app.settings.lastExportFolder)
        # Add a shortcut for the project folder (if saved)
        if self.project.uri:
            shortcut = os.path.dirname(self.project.uri)
            chooser.add_shortcut_folder_uri(shortcut)

        response = chooser.run()
        if response == Gtk.ResponseType.DELETE_EVENT:
            # This happens because Gtk.FileChooserNative is confused because we
            # added a filter but since it's ignored it complains there is none.
            # Try again without the filter.
            chooser.remove_filter(file_filter)
            response = chooser.run()

        if response == Gtk.ResponseType.ACCEPT:
            self.app.settings.lastExportFolder = chooser.get_current_folder()
            self.fileentry.set_text(os.path.join(chooser.get_filename()))

    def _render_button_clicked_cb(self, unused_button):
        """Starts the rendering process."""
        self.__replace_proxies()
        self.__unset_effect_preview_props()
        filename = os.path.realpath(self.fileentry.get_text())
        self.outfile = Gst.filename_to_uri(filename)
        self.app.settings.lastExportFolder = os.path.dirname(filename)
        self.progress = RenderingProgressDialog(self.app, self)
        # Hide the rendering settings dialog while rendering
        self.window.hide()

        self.app.gui.editor.timeline_ui.timeline.set_best_zoom_ratio(allow_zoom_in=True)
        self.project.set_rendering(True)
        self._pipeline.set_render_settings(
            self.outfile, self.project.container_profile)
        self.start_action()
        self.progress.window.show()
        self.progress.connect("cancel", self._cancel_render)
        self.progress.connect("pause", self._pause_render)
        bus = self._pipeline.get_bus()
        bus.add_signal_watch()
        self._gst_signal_handlers_ids[bus] = bus.connect("message", self._bus_message_cb)
        self.project.pipeline.connect("position", self._update_position_cb)
        # Force writing the config now, or the path will be reset
        # if the user opens the rendering dialog again
        self.app.settings.store_settings()

    def _close_button_clicked_cb(self, unused_button):
        self.debug("Render dialog's Close button clicked")
        self.project.disconnect_by_func(self._rendering_settings_changed_cb)
        self.destroy()

    def _delete_event_cb(self, unused_window, unused_event):
        self.debug("Render dialog is being deleted")
        self.destroy()

    def _container_context_help_clicked_cb(self, unused_button):
        show_user_manual("codecscontainers")

    def _fileentry_changed_cb(self, unused_entry):
        self._check_filename()

    # Periodic (timer) callbacks
    def _update_time_estimate_cb(self):
        if self._rendering_is_paused:
            # Do nothing until we resume rendering
            return True

        if self._is_rendering:
            if self.current_position:
                timediff = time.time() - self._time_started - self._time_spent_paused
                length = self.project.ges_timeline.props.duration
                estimated_time = timediff * length / self.current_position
                remaining_time = estimated_time - timediff
                estimate = beautify_eta(int(remaining_time * Gst.SECOND))
                if estimate:
                    self.progress.update_progressbar_eta(estimate)
            return True

        self._time_estimate_timer = None
        self.debug("Stopping the ETA timer")
        return False

    def _update_filesize_estimate_cb(self):
        if self._rendering_is_paused:
            return True  # Do nothing until we resume rendering
        elif self._is_rendering:
            est_filesize = self._get_filesize_estimate()
            if est_filesize:
                self.progress.set_filesize_estimate(est_filesize)
            return True
        else:
            self.debug("Stopping the filesize estimation timer")
            self._filesize_estimate_timer = None
            return False  # Stop the timer

    # GStreamer callbacks
    def _bus_message_cb(self, unused_bus, message):
        if message.type == Gst.MessageType.EOS:  # Render complete
            self.debug("got EOS message, render complete")
            self._shut_down()
            self.progress.progressbar.set_fraction(1.0)
            self.progress.progressbar.set_text(_("Render complete"))
            self.progress.window.set_title(_("Render complete"))
            self.progress.set_filesize_estimate(None)
            if not self.progress.window.is_active():
                notification = _(
                    '"%s" has finished rendering.') % self.fileentry.get_text()
                self.notification = self.app.system.desktop_message(
                    _("Render complete"), notification, "pitivi")
            self._maybe_play_finished_sound()
            self.progress.play_rendered_file_button.show()
            self.progress.close_button.show()
            self.progress.show_in_file_manager_button.show()
            self.progress.cancel_button.hide()
            self.progress.play_pause_button.hide()

        elif message.type == Gst.MessageType.ERROR:
            # Errors in a GStreamer pipeline are fatal. If we encounter one,
            # we should abort and show the error instead of sitting around.
            error, details = message.parse_error()
            self._cancel_render()
            self._show_render_error_dialog(error, details)

        elif message.type == Gst.MessageType.STATE_CHANGED and self.progress:
            if message.src == self._pipeline:
                unused_prev, state, pending = message.parse_state_changed()
                if pending == Gst.State.VOID_PENDING:
                    # State will not change further.
                    if state == Gst.State.PLAYING:
                        self.debug("Inhibiting sleep when rendering")
                        self.app.simple_inhibit(RenderDialog.INHIBIT_REASON,
                                                Gtk.ApplicationInhibitFlags.SUSPEND)
                    else:
                        self.app.simple_uninhibit(RenderDialog.INHIBIT_REASON)

    def _update_position_cb(self, unused_pipeline, position):
        """Updates the progress bar and triggers the update of the file size.

        This one occurs every time the pipeline emits a position changed signal,
        which is *very* often.
        """
        self.current_position = position
        if not self.progress or not position:
            return

        length = self.project.ges_timeline.props.duration
        fraction = float(min(position, length)) / float(length)
        self.progress.update_position(fraction)

        # In order to have enough averaging, only display the ETA after 5s
        timediff = time.time() - self._time_started
        if not self._time_estimate_timer:
            if timediff < 6:
                self.progress.progressbar.set_text(_("Estimating..."))
            else:
                self._time_estimate_timer = GLib.timeout_add_seconds(
                    3, self._update_time_estimate_cb)

        # Filesize is trickier and needs more time to be meaningful.
        if not self._filesize_estimate_timer and (fraction > 0.33 or timediff > 180):
            self._filesize_estimate_timer = GLib.timeout_add_seconds(
                5, self._update_filesize_estimate_cb)

    def __element_added_cb(self, unused_bin, gst_element):
        self.__set_properties(gst_element)

    def __set_properties(self, gst_element):
        """Sets properties on the specified Gst.Element."""
        factory = gst_element.get_factory()
        settings = {}
        if factory == get_combo_value(self.video_encoder_combo):
            settings = self.project.vcodecsettings
        elif factory == get_combo_value(self.audio_encoder_combo):
            settings = self.project.acodecsettings

        for propname, value in settings.items():
            gst_element.set_property(propname, value)
            self.debug("Setting %s to %s", propname, value)

    # Settings changed callbacks
    def _scale_spinbutton_changed_cb(self, unused_button):
        render_scale = self.scale_spinbutton.get_value()
        self.project.render_scale = render_scale
        self.update_resolution()

    def update_resolution(self):
        width, height = self.project.get_video_width_and_height(render=True)
        self.resolution_label.set_text("%d×%d" % (width, height))

    def _project_settings_button_clicked_cb(self, unused_button):
        dialog = ProjectSettingsDialog(self.window, self.project, self.app)
        dialog.window.run()
        self._display_settings()

    def _audio_output_checkbutton_toggled_cb(self, unused_audio):
        self._update_audio_widgets_sensitivity()

    def _video_output_checkbutton_toggled_cb(self, unused_video):
        self._update_video_widgets_sensitivity()

    def __update_render_button_sensitivity(self):
        video_enabled = self.video_output_checkbutton.get_active()
        audio_enabled = self.audio_output_checkbutton.get_active()
        self.render_button.set_sensitive(self._is_filename_valid and
                                         (video_enabled or audio_enabled))

    def _frame_rate_combo_changed_cb(self, combo):
        if self._setting_encoding_profile:
            return

        framerate = get_combo_value(combo)
        self.project.videorate = framerate

    def _video_encoder_combo_changed_cb(self, combo):
        if self._setting_encoding_profile:
            return

        factory = get_combo_value(combo)
        name = factory.get_name()
        self.project.vencoder = name
        if not self.muxer_combo_changing:
            # The user directly changed the video encoder combo.
            self.debug("User chose a video encoder: %s", name)
            self.preferred_vencoder = name
        self._update_valid_video_restrictions(factory)
        self._update_quality_scale()

    def _video_settings_button_clicked_cb(self, unused_button):
        if self._setting_encoding_profile:
            return

        factory = get_combo_value(self.video_encoder_combo)
        self._element_settings_dialog(factory, "video")

    def _channels_combo_changed_cb(self, combo):
        if self._setting_encoding_profile:
            return

        self.project.audiochannels = get_combo_value(combo)

    def _sample_rate_combo_changed_cb(self, combo):
        if self._setting_encoding_profile:
            return

        self.project.audiorate = get_combo_value(combo)

    def _audio_encoder_changed_combo_cb(self, combo):
        if self._setting_encoding_profile:
            return

        factory = get_combo_value(combo)
        name = factory.get_name()
        self.project.aencoder = name
        if not self.muxer_combo_changing:
            # The user directly changed the audio encoder combo.
            self.debug("User chose an audio encoder: %s", name)
            self.preferred_aencoder = name
        self._update_valid_audio_restrictions(factory)

    def _audio_settings_button_clicked_cb(self, unused_button):
        factory = get_combo_value(self.audio_encoder_combo)
        self._element_settings_dialog(factory, "audio")

    def _update_file_extension(self):
        # Update the extension of the filename.
        basename = os.path.splitext(self.fileentry.get_text())[0]
        self._update_filename(basename)

    def _muxer_combo_changed_cb(self, combo):
        """Handles the changing of the container format combobox."""
        if self._setting_encoding_profile:
            return

        factory = get_combo_value(combo)
        self.project.muxer = factory.get_name()

        self._update_file_extension()

        # Update muxer-dependent widgets.
        self.update_available_encoders()

    def _update_preset_selection_menubutton_func(self, source_widget, target_widget):
        if self._setting_encoding_profile:
            return

        preset_item = self.presets_manager.matching_preset()
        self.presets_manager.select_preset(preset_item)

    def _update_quality_scale(self):
        encoder = get_combo_value(self.video_encoder_combo)
        adapter = quality_adapters.get(encoder.get_name())

        self.quality_scale.set_sensitive(bool(adapter))

        if adapter:
            quality = adapter.calculate_quality(self.project.vcodecsettings)
        else:
            quality = self.quality_adjustment.props.lower
        self.quality_adjustment.handler_block(self.quality_adjustment_handler_id)
        try:
            self.quality_adjustment.props.value = quality
        finally:
            self.quality_adjustment.handler_unblock(self.quality_adjustment_handler_id)

    def _quality_adjustment_value_changed_cb(self, adjustment):
        self.apply_vcodecsettings_quality(self.quality_adjustment.props.value)

    def apply_vcodecsettings_quality(self, quality):
        encoder = get_combo_value(self.video_encoder_combo)
        adapter = quality_adapters.get(encoder.get_name())
        if not adapter:
            # The current video encoder is not yet supported.
            return

        adapter.update_project_vcodecsettings(self.project, round(quality))



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2013, 2014, 2015, Thibault Saunier <tsaunier@gnome.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
"""Project settings dialog."""
import datetime
import os

from gi.repository import Gst
from gi.repository import Gtk

from pitivi.configure import get_ui_dir
from pitivi.dialogs.prefs import PreferencesDialog
from pitivi.preset import VideoPresetManager
from pitivi.utils.ripple_update_group import RippleUpdateGroup
from pitivi.utils.ui import create_frame_rates_model
from pitivi.utils.ui import get_combo_value
from pitivi.utils.ui import set_combo_value
from pitivi.utils.widgets import FractionWidget


class ProjectSettingsDialog:
    """Manager of a dialog for viewing and changing the project settings.

    Attributes:
        project (Project): The project who's settings are displayed.
        app (Pitivi): The current app.
    """

    def __init__(self, parent_window, project, app):
        self.app = app
        self.project = project
        self.video_presets = VideoPresetManager(app.system)

        self.sar = 0
        self.proxy_aspect_ratio = Gst.Fraction(1, 0)

        self._create_ui()
        self.window.set_transient_for(parent_window)
        self._setup_ui_constraints()
        self.update_ui()

    def _create_ui(self):
        """Initializes the static parts of the UI."""
        self.builder = Gtk.Builder()
        self.builder.add_from_file(
            os.path.join(get_ui_dir(), "projectsettings.ui"))
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("project-settings-dialog")
        self.frame_rate_combo = self.builder.get_object("frame_rate_combo")
        self.year_spinbutton = self.builder.get_object("year_spinbutton")
        self.author_entry = self.builder.get_object("author_entry")
        self.width_spinbutton = self.builder.get_object("width_spinbutton")
        self.height_spinbutton = self.builder.get_object("height_spinbutton")
        self.video_presets_combo = self.builder.get_object("video_presets_combo")
        self.constrain_sar_button = self.builder.get_object("constrain_sar_button")
        self.select_dar_radiobutton = self.builder.get_object("select_dar_radiobutton")
        self.year_spinbutton = self.builder.get_object("year_spinbutton")

        self.video_preset_menubutton = self.builder.get_object("video_preset_menubutton")
        self.video_presets.setup_ui(self.video_presets_combo,
                                    self.video_preset_menubutton)
        self.video_presets.connect("preset-loaded", self.__video_preset_loaded_cb)

        self.scaled_proxy_width_spin = self.builder.get_object("scaled_proxy_width")
        self.scaled_proxy_height_spin = self.builder.get_object("scaled_proxy_height")
        self.proxy_res_linked_check = self.builder.get_object("proxy_res_linked")

        self.title_horizontal_spinbutton = self.builder.get_object("title_safe_area_horizontal")
        self.title_vertical_spinbutton = self.builder.get_object("title_safe_area_vertical")
        self.action_horizontal_spinbutton = self.builder.get_object("action_safe_area_horizontal")
        self.action_vertical_spinbutton = self.builder.get_object("action_safe_area_vertical")

    def _setup_ui_constraints(self):
        """Creates the dynamic widgets and connects other widgets."""
        # Add custom framerate fraction widget.
        frame_rate_box = self.builder.get_object("frame_rate_box")
        self.frame_rate_fraction_widget = FractionWidget()
        frame_rate_box.pack_end(self.frame_rate_fraction_widget, True, True, 0)
        self.frame_rate_fraction_widget.show()

        # Behavior.
        self.widgets_group = RippleUpdateGroup()
        self.widgets_group.add_vertex(self.frame_rate_combo,
                                      signal="changed",
                                      update_func=self._update_frame_rate_combo_func,
                                      update_func_args=(self.frame_rate_fraction_widget,))
        self.widgets_group.add_vertex(self.frame_rate_fraction_widget,
                                      signal="value-changed",
                                      update_func=self._update_frame_rate_fraction_func,
                                      update_func_args=(self.frame_rate_combo,))
        self.widgets_group.add_vertex(self.width_spinbutton, signal="value-changed")
        self.widgets_group.add_vertex(self.height_spinbutton, signal="value-changed")
        self.widgets_group.add_vertex(self.video_preset_menubutton,
                                      update_func=self._update_preset_menu_button_func,
                                      update_func_args=(self.video_presets,))
        self.widgets_group.add_vertex(self.scaled_proxy_width_spin, signal="value-changed")
        self.widgets_group.add_vertex(self.scaled_proxy_height_spin, signal="value-changed")

        # Constrain width and height IFF the Constrain checkbox is checked.
        # Video
        self.widgets_group.add_edge(self.width_spinbutton, self.height_spinbutton,
                                    predicate=self.width_height_linked,
                                    edge_func=self.update_height)
        self.widgets_group.add_edge(self.height_spinbutton, self.width_spinbutton,
                                    predicate=self.width_height_linked,
                                    edge_func=self.update_width)
        # Proxy
        self.widgets_group.add_edge(self.scaled_proxy_width_spin,
                                    self.scaled_proxy_height_spin,
                                    predicate=self.proxy_res_linked,
                                    edge_func=self.update_scaled_proxy_height)
        self.widgets_group.add_edge(self.scaled_proxy_height_spin,
                                    self.scaled_proxy_width_spin,
                                    predicate=self.proxy_res_linked,
                                    edge_func=self.update_scaled_proxy_width)

        # Keep the framerate combo and fraction widgets in sync.
        self.widgets_group.add_bi_edge(
            self.frame_rate_combo, self.frame_rate_fraction_widget)

        # Presets.
        self.video_presets.load_all()

        # Bind the widgets in the Video tab to the Video Presets Manager.
        self.bind_spinbutton(self.video_presets, "width", self.width_spinbutton)
        self.bind_spinbutton(self.video_presets, "height", self.height_spinbutton)
        self.bind_fraction_widget(
            self.video_presets, "frame-rate", self.frame_rate_fraction_widget)

        self.widgets_group.add_edge(
            self.frame_rate_fraction_widget, self.video_preset_menubutton)
        self.widgets_group.add_edge(self.width_spinbutton, self.video_preset_menubutton)
        self.widgets_group.add_edge(self.height_spinbutton, self.video_preset_menubutton)

    def bind_fraction_widget(self, mgr, name, widget):
        mgr.bind_widget(name, widget.set_widget_value, widget.get_widget_value)

    def bind_combo(self, mgr, name, widget):
        def setter(value):
            res = set_combo_value(widget, value)
            assert res, value
        mgr.bind_widget(name, setter, lambda: get_combo_value(widget))

    def bind_spinbutton(self, mgr, name, widget):
        mgr.bind_widget(name,
                        lambda x: widget.set_value(float(x)),
                        lambda: int(widget.get_value()))

    def width_height_linked(self):
        return self.constrain_sar_button.props.active and not self.video_presets.ignore_update_requests

    def proxy_res_linked(self):
        return self.proxy_res_linked_check.props.active

    def _update_frame_rate_fraction_func(self, unused, fraction_widget, combo_widget):
        """Updates the fraction_widget to match the combo_widget."""
        fraction_widget.set_widget_value(get_combo_value(combo_widget))

    def _update_frame_rate_combo_func(self, unused, combo_widget, fraction_widget):
        """Updates the combo_widget to match the fraction_widget."""
        widget_value = fraction_widget.get_widget_value()
        fr_datum = (widget_value.num, widget_value.denom)
        model = create_frame_rates_model(fr_datum)
        self.frame_rate_combo.set_model(model)
        res = set_combo_value(combo_widget, widget_value)
        assert res, widget_value

    def __video_preset_loaded_cb(self, unused_mgr):
        self.sar = self.get_sar()

    def get_sar(self):
        width = int(self.width_spinbutton.get_value())
        height = int(self.height_spinbutton.get_value())
        return Gst.Fraction(width, height)

    def _constrain_sar_button_toggled_cb(self, unused_button):
        self.sar = self.get_sar()

    def _update_preset_menu_button_func(self, unused_source, unused_target, mgr):
        mgr.update_menu_actions()

    def update_width(self):
        height = int(self.height_spinbutton.get_value())
        fraction = height * self.sar
        width = int(fraction.num / fraction.denom)
        self.width_spinbutton.set_value(width)

    def update_height(self):
        width = int(self.width_spinbutton.get_value())
        fraction = width / self.sar
        height = int(fraction.num / fraction.denom)
        self.height_spinbutton.set_value(height)

    def _proxy_res_linked_toggle_cb(self, unused_button):
        width = int(self.scaled_proxy_width_spin.get_value())
        height = int(self.scaled_proxy_height_spin.get_value())
        self.proxy_aspect_ratio = Gst.Fraction(width, height)

    def _proxy_settings_label_cb(self, unused_widget, unused_parm):
        prefs_dialog = PreferencesDialog(self.app)
        prefs_dialog.stack.set_visible_child_name("_proxies")
        prefs_dialog.run()

    def update_scaled_proxy_width(self):
        height = int(self.scaled_proxy_height_spin.get_value())
        fraction = height * self.proxy_aspect_ratio
        width = int(fraction.num / fraction.denom)
        self.scaled_proxy_width_spin.set_value(width)

    def update_scaled_proxy_height(self):
        width = int(self.scaled_proxy_width_spin.get_value())
        fraction = width / self.proxy_aspect_ratio
        height = int(fraction.num / fraction.denom)
        self.scaled_proxy_height_spin.set_value(height)

    def update_ui(self):
        # Video
        self.width_spinbutton.set_value(self.project.videowidth)
        self.height_spinbutton.set_value(self.project.videoheight)
        self.frame_rate_fraction_widget.set_widget_value(self.project.videorate)

        matching_video_preset = self.video_presets.matching_preset(self.project)
        if matching_video_preset:
            self.video_presets_combo.set_active_id(matching_video_preset)

        # Safe Areas
        self.title_vertical_spinbutton.set_value(self.project.title_safe_area_vertical * 100)
        self.title_horizontal_spinbutton.set_value(self.project.title_safe_area_horizontal * 100)
        self.action_vertical_spinbutton.set_value(self.project.action_safe_area_vertical * 100)
        self.action_horizontal_spinbutton.set_value(self.project.action_safe_area_horizontal * 100)

        # Metadata
        self.author_entry.set_text(self.project.author)
        if self.project.year:
            year = int(self.project.year)
        else:
            year = datetime.datetime.now().year
        self.year_spinbutton.get_adjustment().set_value(year)

        self.scaled_proxy_width_spin.set_value(self.project.scaled_proxy_width)
        self.scaled_proxy_height_spin.set_value(self.project.scaled_proxy_height)

    def update_project(self):
        with self.app.action_log.started("change project settings",
                                         toplevel=True):
            self.project.author = self.author_entry.get_text()
            self.project.year = str(self.year_spinbutton.get_value_as_int())

            self.project.set_video_properties(
                int(self.width_spinbutton.get_value()),
                int(self.height_spinbutton.get_value()),
                self.frame_rate_fraction_widget.get_widget_value())

            # Store values as a decimal value
            self.project.set_safe_areas_sizes(int(self.title_horizontal_spinbutton.get_value()) / 100,
                                              int(self.title_vertical_spinbutton.get_value()) / 100,
                                              int(self.action_horizontal_spinbutton.get_value()) / 100,
                                              int(self.action_vertical_spinbutton.get_value()) / 100)

            proxy_width = int(self.scaled_proxy_width_spin.get_value())
            proxy_height = int(self.scaled_proxy_height_spin.get_value())
            # Update scaled proxy meta-data and trigger proxy regen
            if not self.project.has_scaled_proxy_size() or \
                    self.project.scaled_proxy_width != proxy_width or \
                    self.project.scaled_proxy_height != proxy_height:
                self.project.scaled_proxy_width = proxy_width
                self.project.scaled_proxy_height = proxy_height

                self.project.regenerate_scaled_proxies()

    def _response_cb(self, unused_widget, response):
        """Handles the dialog being closed."""
        if response == Gtk.ResponseType.OK:
            self.update_project()
        self.window.destroy()



# -*- coding: utf-8 -*-
# Pitivi video editor
# Copyright (c) 2020, Jackson Eickhoff <jacksoneick@gmail.com>
# Copyright (c) 2020, Tanner Skelton <tskelton@huskers.unl.edu>
# Copyright (c) 2020, Cordell Rhoads <rhoadscordell7@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, see <http://www.gnu.org/licenses/>.
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

from pitivi.utils.loggable import Loggable


class AlignmentEditor(Gtk.EventBox, Loggable):
    """Widget for aligning a video clip.

    Attributes:
        app (Pitivi): The app.
        _project (Project): The project.
    """

    __gtype_name__ = "AlignmentEditor"

    __gsignals__ = {
        "align": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        Gtk.EventBox.__init__(self)
        Loggable.__init__(self)
        self._hovered_box = None

        self.connect("button-release-event", self._button_release_event_cb)
        self.connect("motion-notify-event", self._motion_notify_event_cb)
        self.connect("leave-notify-event", self._leave_notify_event_cb)
        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK)

    def _leave_notify_event_cb(self, unused_widget, _):
        self._hovered_box = None
        self.queue_draw()

    def _button_release_event_cb(self, widget, event):
        if not self._hovered_box:
            return
        self.emit("align")

    def get_clip_position(self, project, source):
        """Returns corresponding clip position in the viewer."""
        clip_width = source.get_child_property("width").value
        clip_height = source.get_child_property("height").value
        x = self.__calculate_clip_position(self._hovered_box[0], clip_width, project.videowidth)
        y = self.__calculate_clip_position(self._hovered_box[1], clip_height, project.videoheight)

        return x, y

    def __calculate_clip_position(self, index, clip_size, project_size):
        if index == 0:
            coordinate = -clip_size
        elif index == 1:
            coordinate = -clip_size / 2
        elif index == 2:
            coordinate = 0
        elif index == 3:
            coordinate = project_size / 2 - clip_size / 2
        elif index == 4:
            coordinate = project_size - clip_size
        elif index == 5:
            coordinate = project_size - clip_size / 2
        elif index == 6:
            coordinate = project_size
        else:
            coordinate = 0

        return int(coordinate)

    def _motion_notify_event_cb(self, widget, event):
        hovered_box = self._get_box(event.x, event.y)
        if hovered_box != self._hovered_box:
            self._hovered_box = hovered_box
            self.queue_draw()

    def get_used_size(self):
        """Returns the size used for drawing.

        For drawing pixel perfect lines, we need to be able to divide
        the width and height between four boxes and three 1px lines,
        such that the box size is an int value:

            box_size + 1 + box_size + 1 + box_size + 1 + box_size
        """
        box_width = ((self.get_allocated_width() - 3) // 4)
        box_height = ((self.get_allocated_height() - 3) // 4)
        width = box_width * 4 + 3
        height = box_height * 4 + 3
        return width, height, box_width, box_height

    def do_draw(self, cr):
        width, height, box_width, box_height = self.get_used_size()

        self._draw_frame(cr, width, height)

        # Highlight the box that the cursor is hovering over
        if self._hovered_box:
            x = width / 8 * self._hovered_box[0]
            y = height / 8 * self._hovered_box[1]
            cr.rectangle(x, y, box_width, box_height)

            color = self.get_style_context().get_color(Gtk.StateFlags.LINK)
            cr.set_source_rgba(color.red, color.green, color.blue, 0.7)
            cr.set_line_width(1)
            cr.fill()
            cr.stroke()

    def _draw_frame(self, cr, width, height):
        # How far the line should be displaced from the edge of the widget
        line_offset_x = (width - 3) / 4 + 0.5
        line_offset_y = (height - 3) / 4 + 0.5

        cr.move_to(0, line_offset_y)
        cr.line_to(width, line_offset_y)

        cr.move_to(0, height - line_offset_y)
        cr.line_to(width, height - line_offset_y)

        cr.move_to(line_offset_x, 0)
        cr.line_to(line_offset_x, height)

        cr.move_to(width - line_offset_x, 0)
        cr.line_to(width - line_offset_x, height)

        color = self.get_style_context().get_color(Gtk.StateFlags.ACTIVE)
        cr.set_source_rgba(color.red, color.green, color.blue, 0.6)
        cr.set_line_width(1)
        cr.stroke()

        # cr.rectangle(line_offset_x + 0.5, line_offset_y + 0.5, width - line_offset_x * 2 - 1, height - line_offset_y * 2 - 1)
        cr.rectangle(line_offset_x, line_offset_y, width - line_offset_x * 2, height - line_offset_y * 2)

        cr.set_source_rgba(color.red, color.green, color.blue, color.alpha)
        cr.set_line_width(2)
        cr.stroke()

    def _get_box(self, x, y):
        """Returns the box containing the specified pixel."""
        width, height, _, _ = self.get_used_size()

        box_x = int(x * 7 / width)
        box_y = int(y * 7 / height)

        if box_x < 0 or box_x > 6 or box_y < 0 or box_y > 6:
            return None

        return box_x, box_y
